#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for core classes.
"""

import unittest
import sasoptpy as so


class TestVariable(unittest.TestCase):
    """
    Unit tests for :class:`sasoptpy.Variable` objects
    """

    def setUp(self):
        pass

    def test_repr(self):
        x = so.Variable(name='x')
        self.assertEqual(repr(x), "sasoptpy.Variable(name='x', vartype='CONT')")

        y = so.Variable(name='y', lb=1, ub=10, init=3)
        self.assertEqual(repr(y), "sasoptpy.Variable(name='y', lb=1, ub=10, init=3, vartype='CONT')")

        I = so.Set(name='I')
        z = so.VariableGroup(I, name='z')
        self.assertEqual(repr(z[0]), "sasoptpy.Variable(name='z[0]', vartype='CONT')")

    def test_string(self):
        x = so.VariableGroup([1,2,3], name='x')
        self.assertEqual(str(x[1]), "x[1]")

    def test_definition(self):
        x = so.Variable(name='x')
        self.assertEqual(x._defn(), "var x;")

        y = so.Variable(name='y', lb=1, ub=10, init=3, vartype=so.INT)
        self.assertEqual(y._defn(), "var y integer >= 1 <= 10 init 3;")

        z = so.Variable(name='z', ub=0.5, vartype=so.BIN)
        self.assertEqual(z._defn(), "var z binary <= 0.5;")

    def tearDown(self):
        so.reset()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from inspect import cleandoc
import unittest
import os
import sasoptpy as so
import saspy


current_dir = os.path.dirname(os.path.abspath(__file__))


class TestSASInterface(unittest.TestCase):
    """
    Unit tests for the SAS interface
    """

    @classmethod
    def setUpClass(cls):
        raise unittest.SkipTest('Disabling saspy tests for alpha release')
        cls.conn = None
        try:
            cls.conn = None
            if os.name == 'nt':
                cls.conn = saspy.SASsession(cfgname='winlocal')
            else:
                cfg_file = os.path.join(
                    current_dir, '../examples/saspy_config.py')
                cls.conn = saspy.SASsession(cfgfile=cfg_file)
            print('Connected to SAS')
        except TypeError:
            raise unittest.SkipTest('Environment variable may not be defined')

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.endsas()
            cls.conn = None

    def setUp(self):
        so.reset()

    def test_long_names(self):
        if TestSASInterface.conn is None:
            self.skipTest('SAS session is not available')

        vs = 10
        m = so.Model(name='test_long_name', session=TestSASInterface.conn)
        x = m.add_variables(
            vs, name='averylongvariablenamethatdoesnotwork', ub=2, lb=0)
        c = m.add_constraint(
            so.expr_sum(i * x[i] for i in range(vs)) <= 1, name='c')
        o = m.set_objective(
            so.expr_sum(x[i] for i in range(vs)), sense=so.MAX, name='obj')

        def raise_runtime():
            m.solve()

        self.assertRaises(RuntimeError, raise_runtime)

        m.solve(limit_names=True)
        m.get_solution()
        self.assertEqual(x[0].get_value(), 2)

        m.set_objective(x[0], sense=so.MIN, name='obj2')
        m.solve(limit_names=True, submit=False)
        self.assertEqual(x[0].get_value(), 2)

    def test_long_lines(self):
        if TestSASInterface.conn is None:
            self.skipTest('SAS session is not available')

        vs = 2000
        m = so.Model(name='test_long_line', session=TestSASInterface.conn)
        x = m.add_variables(vs, name='averylongvariablename', ub=2, lb=0)
        c = m.add_constraint(
            so.expr_sum(i * x[i] for i in range(vs)) <= 1, name='c')
        o = m.set_objective(x[0], sense=so.MAX, name='obj')

        m.solve(wrap_lines=True)

        self.assertEqual(x[0].get_value(), 2)

    def test_mps_on_sas(self):
        if TestSASInterface.conn is None:
            self.skipTest('SAS session is not available')

        m = so.Model(name='test_mps_on_sas', session=TestSASInterface.conn)
        x = m.add_variables(10, name='x', lb=0)
        m.add_constraints((x[i] >= i for i in range(10)), name='c')
        m.set_objective(
            so.expr_sum(x[i] for i in range(10)), sense=so.MIN, name='obj')

        m.solve(mps=True, verbose=True)
        self.assertEqual(x[2].get_value(), 2)

    def test_workspace_post_value(self):

        if TestSASInterface.conn is None:
            self.skipTest('SAS session is not available')

        with so.Workspace('test_ws_var_value',
                          session=TestSASInterface.conn) as w:
            x = so.Variable(name='x')
            c = so.Constraint(x <= 5, name='c')
            o = so.Objective(x, sense=so.MAX, name='obj')
            so.actions.solve()
        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x;
               con c : x <= 5;
               max obj = x;
               solve;
            quit;'''))
        w.submit()
        self.assertEqual(x.get_value(), 5)

    def test_forced_optmodel(self):
        if not TestSASInterface.conn:
            self.skipTest('No session is available')

        # Forced to OPTMODEL due abstract elements
        m = so.Model(name='test_forced_optmodel', session=TestSASInterface.conn)
        S = so.Set(name='S')
        m.include(S)
        x = m.add_variable(name='x', ub=1)
        m.set_objective(2*x, sense=so.MAX, name='obj')
        def warn_abstract():
            response = m.solve(mps=True, submit=False)
            self.assertIn('set S;', response)
        self.assertWarns(UserWarning, warn_abstract)

        # Uploaded MPS Table
        m.drop(S)
        response = m.solve(mps=True, submit=False)

        # Forced to OPTMODEL due nonlinearity
        m.set_objective(x**2, sense=so.MAX, name='obj2')
        def warn_nonlinear():
            response = m.solve(mps=True, submit=False)
            self.assertIn('var x', response)
        self.assertWarns(UserWarning, warn_nonlinear)

        # Forced to give error
        def force_decomp():
            m.solve(mps=True, options={'decomp': 'USER'})
        self.assertRaises(RuntimeError, force_decomp)

        # No objective
        m.solve(mps=True, verbose=True, name='no_obj')

    def test_postprocess_optmodel_string(self):

        if TestSASInterface.conn is None:
            self.skipTest('SAS session is not available')

        with so.Workspace('w', session=TestSASInterface.conn) as w:
            x = so.Variable(name='x')
            o = so.Objective((x-1)**2, sense=so.MIN, name='obj')
            so.actions.solve()
        w.submit(limit_names=True, wrap_lines=True)
        self.assertEqual(x.get_value(), 1)

    def test_optmodel_error(self):

        if TestSASInterface.conn is None:
            self.skipTest('SAS session is not available')

        with so.Workspace('w', session=TestSASInterface.conn) as w:
            so.abstract.LiteralStatement('abc')
        def produce_error():
            w.submit(verbose=True)
        self.assertRaises(RuntimeError, produce_error)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for set iterators.
"""

import os
import sys
import unittest
import warnings
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '..')))
from util import assert_equal_wo_temps

from tests.swat_config import create_cas_connection


class TestSetIterator(unittest.TestCase):
    """
    Unit tests for :class:`sasoptpy.abstract.SetIterator` objects
    """

    @classmethod
    def setUpClass(cls):
        so.reset()
        cls.conn = None
        from swat import CAS, SWATError
        try:
            cls.conn = create_cas_connection()
        except SWATError:
            warnings.warn('CAS connection is not available', RuntimeWarning)
        except TypeError:
            warnings.warn('CAS variables are not available', RuntimeWarning)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def test_set_iterator_init(self):
        S = so.Set(name='S')
        i = so.SetIterator(S, name='i')

        self.assertEqual(i.get_name(), 'i')
        self.assertEqual(i.get_type(), so.NUM)

        for j in S:
            assert_equal_wo_temps(self, j._get_for_expr(), 'o1 in S')
            assert_equal_wo_temps(
                self, repr(j),
                'sasoptpy.SetIterator(S, name=\'o1\')'
            )

    def test_set_iterator_as_exp(self):
        S = so.Set(name='S')
        x = so.VariableGroup(S, name='x')
        for i in S:
            e = i + x[0]
            assert_equal_wo_temps(self, so.to_expression(e), 'o1 + x[0]')
            assert_equal_wo_temps(self, so.to_definition(i), 'o1 in S')

    def test_set_iterator_condition(self):

        from sasoptpy.actions import condition

        S = so.Set(name='S')
        P = so.Set(name='P')
        x = so.VariableGroup(S, name='x')

        # GT
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(2*i > 1):
                c[i] = i * x[i] <= 5

        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c {o4 in S: 2.0 * o4 > 1} : o4 * x[o4] <= 5;')

        # GE
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(2 * i >= 1):
                c[i] = i * x[i] <= 5

        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c {o4 in S: 2.0 * o4 >= 1} : o4 * x[o4] <= 5;')

        # LT
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(2 * i < 1):
                c[i] = i * x[i] <= 5

        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c {o4 in S: 2.0 * o4 < 1} : o4 * x[o4] <= 5;')

        # LE
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(2 * i <= 1):
                c[i] = i * x[i] <= 5

        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c {o4 in S: 2.0 * o4 <= 1} : o4 * x[o4] <= 5;')

        # EQ
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(2 * i == 1):
                c[i] = i * x[i] <= 5

        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c {o4 in S: 2.0 * o4 = 1} : o4 * x[o4] <= 5;')

        # NE
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(2 * i != 1):
                c[i] = i * x[i] <= 5

        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c {o4 in S: 2.0 * o4 ne 1} : o4 * x[o4] <= 5;')

        # IN
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(i.sym in P):
                c[i] = i * x[i] <= 5

        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c {o4 in S: o4 IN P} : o4 * x[o4] <= 5;')

        # Alternatively:
        # assert_equal_wo_temps(
        #     self, so.to_definition(c),
        #     'con c {o4 in S: <o4> IN P} : o4 * x[o4] <= 5;')

        # AND, OR, Multi Statement
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            e1 = 2 * i != 1
            e2 = i ** 2 <= 3
            e3 = 3 * i >= 9
            e4 = i != 0
            with condition(((e1) & (e2) | (e3)) & (e4)):
                c[i] = i * x[i] <= 5

        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c {o34 in S: ((((2.0 * o34 ne 1) and ((o34) ^ (2) <= 3)) or (3.0 * o34 >= 9)) and (o34 NE 0))} : o34 * x[o34] <= 5;'
        )

    def test_set_iterator_group(self):

        S = so.Set(name='S', settype=[so.STR, so.NUM])
        x = so.VariableGroup(S, name='x')
        p = so.ParameterGroup(S, name='p')
        c = so.ConstraintGroup(None, name='c')
        d = so.ConstraintGroup(None, name='d')
        for k in S:
            (i, j) = k
            c[k] = so.Constraint(x[i, j] + 5 <= 3*i - 2*j)
            for t in range(3):
                d[k, t] = so.Constraint(x[i, t] <= 4 * j)

        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c_o7 {<o8, o10> in S} : x[o8, o10] - 3.0 * o8 + 2.0 * o10 <= -5;')
        assert_equal_wo_temps(
            self, so.to_definition(d),
            cleandoc('''
                con d_o7_0 {<o8, o10> in S} : x[o8, 0] - 4.0 * o10 <= 0;
                con d_o7_1 {<o8, o10> in S} : x[o8, 1] - 4.0 * o10 <= 0;
                con d_o7_2 {<o8, o10> in S} : x[o8, 2] - 4.0 * o10 <= 0;
            ''')
        )

    def test_set_iterator_in_iterate(self):

        with so.Workspace('test_set_iter_in_iterate') as w:
            S = so.Set(name='S', settype=[so.NUM, so.NUM])
            W = so.Set(name='W', settype=so.NUM)
            p = so.ParameterGroup(S, W, name='p')
            x = so.VariableGroup(S, name='x')

            from sasoptpy.actions import read_data
            from sasoptpy.util import iterate

            with iterate(S, name=['pos1', 'pos2']) as keys:
                r = read_data(
                    table='cas_table_1',
                    index={
                        'target': S,
                        'key': keys
                    }
                )
                with iterate(W, name='d') as d:
                    r.append({
                        'index': d,
                        'target': p[keys['pos1'], keys['pos2'], d],
                        'column': d
                    })

                assert_equal_wo_temps(self, so.to_definition(keys),
                                      '<pos1, pos2> in S')
                self.assertEqual(repr(keys), cleandoc('''
                    sasoptpy.SetIteratorGroup(S, datatype=['num', 'num'], names=['pos1', 'pos2'])
                '''))
                self.assertEqual(so.to_expression(keys), 'pos1, pos2')
                self.assertEqual(str(keys), '(pos1, pos2)')


        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set <num, num> S;
               set W;
               num p {S, W};
               var x {{S}};
               read data cas_table_1 into S=[pos1 pos2] {d in W} < p[pos1, pos2, d]=col(d) >;
            quit;
            '''))

    def test_definition_multi(self):
        regset = so.Set(name='regset')
        multiset = so.Set(name='multiset', settype=[so.string, so.number])
        mt = so.VariableGroup(5, name='mt')
        e = so.expr_sum(mt[k] for k in regset)
        self.assertEqual(so.to_expression(e), "sum {k in regset} (mt[k])")
        e = so.expr_sum(mt[k] for k in multiset)
        assert_equal_wo_temps(self, so.to_expression(e),
                              "sum {<o1, o2> in multiset} (mt[o1, o2])")

    def tearDown(self):
        pass



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for implicit variables.
"""

import os
import sys
import unittest
import warnings
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '..')))
from util import assert_equal_wo_temps

from tests.swat_config import create_cas_connection



class TestImplicitVariable(unittest.TestCase):
    """
    Unit tests for :class:`sasoptpy.abstract.ImplicitVar` objects
    """

    @classmethod
    def setUpClass(cls):
        so.reset()
        cls.conn = None
        from swat import CAS, SWATError
        try:
            cls.conn = create_cas_connection()
        except SWATError:
            warnings.warn('CAS connection is not available', RuntimeWarning)
        except TypeError:
            warnings.warn('CAS variables are not available', RuntimeWarning)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def setUp(self):
        pass

    def test_impvar_defn(self):
        x = so.ImplicitVar(0, name='x')
        self.assertEqual(str(x), 'x')
        self.assertEqual(x.get_name(), 'x')
        self.assertEqual(so.to_definition(x), 'impvar x = 0;')

    def test_impvar_setitem(self):
        # Regular assignment
        x = so.Variable(name='x')
        y = so.ImplicitVar((i * x for i in range(3)), name='y')
        self.assertEqual(so.to_expression(y[2]), "2 * x")
        self.assertEqual(so.to_definition(y), cleandoc("""
            impvar y_0 = 0;
            impvar y_1 = x;
            impvar y_2 = 2 * x;
            """))

        p = so.Parameter(name='p', value=4)
        z = so.ImplicitVar((i * p for i in so.exp_range(1, 5)), name='z')
        self.assertEqual(so.to_definition(z), cleandoc("""
            impvar z_1 = p;
            impvar z_2 = 2 * p;
            impvar z_3 = 3 * p;
            impvar z_4 = 4 * p;
            """))

        t = so.ImplicitVar(p, name='t')
        self.assertEqual(so.to_definition(t), 'impvar t = p;')

    def test_impvar_combinations(self):
        # Constant
        x = so.ImplicitVar(4, name='x')
        self.assertEqual(so.to_definition(x), 'impvar x = 4;')

        # Single N
        y = so.ImplicitVar((i ** 2 for i in range(3)), name='y')
        self.assertEqual(so.to_definition(y), cleandoc("""
            impvar y_0 = 0;
            impvar y_1 = 1;
            impvar y_2 = 4;
            """))

        # Single S
        so.reset()
        S = so.Set(name='S')
        y = so.ImplicitVar((2 * i for i in S), name='y')
        self.assertEqual(so.to_definition(y), cleandoc("""
            impvar y {o2 in S} = 2.0 * o2;
            """))

        # Double NN
        y = so.ImplicitVar((i + j for i in range(3) for j in range(2)), name='y')
        self.assertEqual(so.to_definition(y), cleandoc("""
            impvar y_0_0 = 0;
            impvar y_0_1 = 1;
            impvar y_1_0 = 1;
            impvar y_1_1 = 2;
            impvar y_2_0 = 2;
            impvar y_2_1 = 3;
            """))

        # Double NS
        so.reset()
        y = so.ImplicitVar((i + j for i in range(3) for j in S), name='y')
        assert_equal_wo_temps(self, so.to_definition(y), cleandoc("""
            impvar y_0 {o1 in S} = o1;
            impvar y_1 {o2 in S} = o2 + 1;
            impvar y_2 {o3 in S} = o3 + 2;
            """))

        # Double SN
        so.reset()
        y = so.ImplicitVar((i + j for i in S for j in range(2)), name='y')
        self.assertEqual(so.to_definition(y), cleandoc("""
            impvar y_0 {o1 in S} = o1;
            impvar y_1 {o1 in S} = o1 + 1;
            """))

        # Double SS
        id = so.itemid
        y = so.ImplicitVar((i + j for i in S for j in S), name='y')
        assert_equal_wo_temps(self, so.to_definition(y), cleandoc("""
            impvar y {o8 in S, o11 in S} = o8 + o11;
            """))

        # Triple NNN
        y = so.ImplicitVar((i + 2*j for i in range(1) for j in range(2) for k in ['A', 'B']), name='y')
        self.assertEqual(so.to_definition(y), cleandoc("""
            impvar y_0_0_A = 0;
            impvar y_0_0_B = 0;
            impvar y_0_1_A = 2;
            impvar y_0_1_B = 2;
            """))

        # Triple NNS
        so.reset()
        id = so.itemid
        y = so.ImplicitVar((i + j + k for i in range(2) for j in range(3) for k in S), name='y')
        assert_equal_wo_temps(self, so.to_definition(y), cleandoc("""
            impvar y_0_0 {o2 in S} = o2;
            impvar y_0_1 {o6 in S} = o6 + 1;
            impvar y_0_2 {o10 in S} = o10 + 2;
            impvar y_1_0 {o14 in S} = o14 + 1;
            impvar y_1_1 {o18 in S} = o18 + 2;
            impvar y_1_2 {o22 in S} = o22 + 3;
            """))

        # Triple NSN
        so.reset()
        y = so.ImplicitVar((i + j + k for i in range(2) for j in S for k in range(3)), name='y')
        assert_equal_wo_temps(self, so.to_definition(y), cleandoc("""
            impvar y_0_0 {o2 in S} = o2;
            impvar y_0_1 {o2 in S} = o2 + 1;
            impvar y_0_2 {o2 in S} = o2 + 2;
            impvar y_1_0 {o10 in S} = o10 + 1;
            impvar y_1_1 {o10 in S} = o10 + 2;
            impvar y_1_2 {o10 in S} = o10 + 3;
            """))

        # Triple NSS
        so.reset()
        y = so.ImplicitVar((i + j + k for i in range(2) for j in S for k in S), name='y')
        assert_equal_wo_temps(self, so.to_definition(y), cleandoc("""
            impvar y_0 {o2 in S, o4 in S} = o2 + o4;
            impvar y_1 {o8 in S, o10 in S} = o8 + o10 + 1;
            """))

        # Triple SNN
        so.reset()
        y = so.ImplicitVar((i + j + k for i in S for j in range(3) for k in range(3)), name='y')
        assert_equal_wo_temps(self, so.to_definition(y), cleandoc("""
            impvar y_0_0 {o1 in S} = o1;
            impvar y_0_1 {o1 in S} = o1 + 1;
            impvar y_0_2 {o1 in S} = o1 + 2;
            impvar y_1_0 {o1 in S} = o1 + 1;
            impvar y_1_1 {o1 in S} = o1 + 2;
            impvar y_1_2 {o1 in S} = o1 + 3;
            impvar y_2_0 {o1 in S} = o1 + 2;
            impvar y_2_1 {o1 in S} = o1 + 3;
            impvar y_2_2 {o1 in S} = o1 + 4;
            """))

        # Triple SNS
        so.reset()
        y = so.ImplicitVar((i + j + k for i in S for j in range(3) for k in S), name='y')
        assert_equal_wo_temps(self, so.to_definition(y), cleandoc("""
            impvar y_0 {o1 in S, o4 in S} = o1 + o4;
            impvar y_1 {o1 in S, o8 in S} = o1 + o8 + 1;
            impvar y_2 {o1 in S, o12 in S} = o1 + o12 + 2;
            """))

        # Triple SSN
        so.reset()
        y = so.ImplicitVar((i + j + k for i in S for j in S for k in range(2)), name='y')
        assert_equal_wo_temps(self, so.to_definition(y), cleandoc("""
            impvar y_0 {o1 in S, o4 in S} = o1 + o4;
            impvar y_1 {o1 in S, o4 in S} = o1 + o4 + 1;
            """))

        # Triple SSS
        so.reset()
        y = so.ImplicitVar((i + j + k + 2 for i in S for j in S for k in S), name='y')
        assert_equal_wo_temps(self, so.to_definition(y), cleandoc("""
            impvar y {o1 in S, o4 in S, o6 in S} = o1 + o4 + o6 + 2;
            """))

    def test_impvar_power(self):
        p = so.Parameter(name='p')
        y = so.ImplicitVar((i + p for i in range(2)), name='y')
        e = y[0] ** 2 + y[1] ** 2
        self.assertEqual(str(e), '(p) ** (2) + (p + 1) ** (2)')
        f = y[0] / p + y[1] / p
        self.assertEqual(str(f), '(p) / (p) + (p + 1) / (p)')

    def test_impvar_shadow(self):
        p = so.Parameter(name='p', value=5)
        x = so.ImplicitVar((p ** i for i in range(3)), name='x')
        S = so.Set(name='S')
        for i in S:
            a = x[i]
            assert_equal_wo_temps(self, str(x[i]), 'x[o13]')

    def test_impvar_keys(self):
        x = so.ImplicitVar(0, name='x')
        self.assertEqual(list(x.get_keys()), [('',)])

        y = so.ImplicitVar((i for i in range(3)), name='y')
        self.assertEqual(list(y.get_keys()), [(0,), (1,), (2,)])

        S = so.Set(name='S')
        z = so.ImplicitVar((2 * i for i in S), name='z')
        first_key = list(z.get_keys())[0][0]
        self.assertTrue(isinstance(first_key, so.SetIterator))

        for i in y:
            self.assertTrue(isinstance(y[i], so.Expression))

    def test_impvar_expr(self):
        x = so.VariableGroup(2, name='x')
        y = so.ImplicitVar((1 - x[i] for i in range(1)), name='y')
        self.assertEqual(so.to_definition(y),
                         'impvar y_0 = - x[0] + 1;')
        e = x[1] + 100 - y[0]
        self.assertEqual(so.to_expression(e),
                         'x[1] - (- x[0] + 1) + 100')

    def tearDown(self):
        pass



from .statement_base import Statement
from .assignment import Assignment
from .cofor_loop import CoForLoopStatement
from .create_data import CreateDataStatement
from .drop_restore import DropStatement, RestoreStatement
from .for_loop import ForLoopStatement
from .if_else import IfElseStatement
from .literal import LiteralStatement
from .objective import ObjectiveStatement
from .read_data import ReadDataStatement
from .solve import SolveStatement
from .fix_unfix import FixStatement, UnfixStatement
from .print import PrintStatement



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Package-wide utility functions
"""

from collections import Iterable
import random
import string
import warnings
from contextlib import contextmanager

import sasoptpy
from sasoptpy.libs import (pd, np)
from threading import RLock


def load_package_globals():
    sasoptpy.itemid = 0

    # Constant values
    sasoptpy.MIN = sasoptpy.minimize = 'MIN'
    sasoptpy.MAX = sasoptpy.maximize = 'MAX'
    sasoptpy.CONT = sasoptpy.continuous = 'CONT'
    sasoptpy.INT = sasoptpy.integer = 'INT'
    sasoptpy.BIN = sasoptpy.binary = 'BIN'
    sasoptpy.STR = sasoptpy.string = 'str'
    sasoptpy.NUM = sasoptpy.number = 'num'

    sasoptpy.LSO = 'lso'
    sasoptpy.BLACKBOX = 'blackbox'
    sasoptpy.MIP = 'milp'
    sasoptpy.LP = 'lp'
    sasoptpy.QP = 'qp'

    sasoptpy.N = sasoptpy.Symbol(name='_N_')

    sasoptpy.ABSTRACT = 'spyABS'
    sasoptpy.CONCRETE = 'spyCON'

    # Transformation dictionary
    sasoptpy._transform = {
        'binary': sasoptpy.BIN,
        'bin': sasoptpy.BIN,
        'integer': sasoptpy.INT,
        'int': sasoptpy.INT,
        'continuous': sasoptpy.CONT,
        'cont': sasoptpy.CONT,
        'maximize': sasoptpy.MAX,
        'max': sasoptpy.MAX,
        'minimize': sasoptpy.MIN,
        'min': sasoptpy.MIN,
        'string': sasoptpy.STR,
        'str': sasoptpy.STR,
        'number': sasoptpy.NUM,
        'num': sasoptpy.NUM
    }

    # Relation dictionary
    sasoptpy._relation_dict = {
        'E': '=',
        'EQ': '=',
        'NE': 'ne',
        'LT': '<',
        'LE': '<=',
        'L': '<=',
        'GT': '>',
        'GE': '>=',
        'G': '>=',
        'IN': 'in'
    }

    sasoptpy.abstract_classes = [
        sasoptpy.Set, sasoptpy.SetIterator, sasoptpy.SetIteratorGroup,
        sasoptpy.Parameter, sasoptpy.ParameterGroup, sasoptpy.ImplicitVar,
        sasoptpy.abstract.ShadowVariable, sasoptpy.abstract.ShadowConstraint,
        sasoptpy.abstract.Statement
        ]

    sasoptpy.loop_operator_tags = ['sum', 'min', 'max']

def get_creation_id():
    sasoptpy.itemid += 1
    return sasoptpy.itemid


def get_next_name():
    return 'o' + str(get_creation_id())


def load_function_containers():
    sasoptpy.container = None
    sasoptpy.container_conditions = False
    sasoptpy.lock = RLock()

    def read_statement_dictionary():
        d = dict()

        d[sasoptpy.core.Model.solve] = \
            sasoptpy.abstract.statement.SolveStatement.solve_model

        d[sasoptpy.core.Variable.set_bounds] = \
            sasoptpy.abstract.statement.Assignment.set_bounds

        d[sasoptpy.abstract.Parameter.set_value] = \
            sasoptpy.abstract.statement.Assignment.set_value

        d[sasoptpy.core.Model.drop_constraint] = \
            sasoptpy.abstract.statement.DropStatement.model_drop_constraint

        d[sasoptpy.core.Model.drop_constraints] = \
            sasoptpy.abstract.statement.DropStatement.model_drop_constraint

        d[sasoptpy.core.Variable.set_value] = \
            sasoptpy.abstract.statement.Assignment.set_value

        return d

    sasoptpy.statement_dictionary = read_statement_dictionary()


def load_default_mediators():
    sasoptpy.mediators['CAS'] = sasoptpy.interface.CASMediator
    sasoptpy.mediators['SAS'] = sasoptpy.interface.SASMediator


def register_to_function_container(func1, func2):
    sasoptpy.statement_dictionary[func1] = func2


def get_in_digit_format(val):
    """
    Returns a value by rounding it to config digits

    Parameters
    ----------
    val : float or integer
        Variable to be formatted into string

    Returns
    -------
    for : str
        Formatted string

    """
    if np.isnan(val):
        return '.'
    digits = sasoptpy.config['max_digits']
    if digits and digits > 0:
        return round(val, digits)
    return val


def get_direction_str(direction):
    return sasoptpy._relation_dict.get(direction)


def _extract_argument_as_list(inp):

    if isinstance(inp, int):
        thelist = list(range(0, inp))
    elif isinstance(inp, range):
        thelist = list(inp)
    elif isinstance(inp, tuple):
        thelist = list(inp)
    elif isinstance(inp, list):
        thelist = inp
    elif sasoptpy.abstract.is_abstract_set(inp):
        thelist = [inp]
    else:
        thelist = list(inp)
    return thelist


def extract_list_value(tuplist, listname):
    """
    Extracts values inside various object types

    Parameters
    ----------
    tuplist : tuple
        Key combination to be extracted
    listname : dict or list or int or float or DataFrame or Series object
        List where the value will be extracted

    """
    if listname is None:
        v = None
    elif isinstance(listname, dict):
        v = listname.get(get_first_member(tuplist), None)
    elif np.isinstance(type(listname), np.number):
        v = listname
    elif isinstance(listname, pd.DataFrame):
        if isinstance(listname.index, pd.MultiIndex):
            v = listname.loc[tuplist[:-1]][tuplist[-1]]
        else:
            v = listname.loc[tuplist]
    elif isinstance(listname, pd.Series):
        v = listname.loc[tuplist]
    else:
        raise ValueError('Parameter type {} is not valid for the operation'.\
                         format(type(listname)))
    return v


def is_key_abstract(key):
    import sasoptpy.abstract
    return sasoptpy.abstract.util.is_key_abstract(key)


def pack_to_tuple(obj):
    """
    Converts a given object to a tuple object

    If the object is a tuple, the function returns the input,
    otherwise creates a single dimensional tuple

    Parameters
    ----------
    obj : Object
        Object that is converted to a tuple

    Returns
    -------
    t : tuple
        Tuple that includes the original object
    """
    if isinstance(obj, tuple):
        return obj
    elif isinstance(obj, str):
        return (obj,)
    return (obj,)


def get_first_member(tp):
    """
    Grabs the first element in a tuple, if a tuple is given as argument

    Parameters
    ----------
    tp : tuple

    """
    if isinstance(tp, tuple):
        if len(tp) == 1:
            return tp[0]
    return tp


def tuple_unpack(tp):
    warnings.warn('Use sasoptpy.util.get_first_member', DeprecationWarning)
    return get_first_member(tp)


def tuple_pack(obj):
    warnings.warn('Use sasoptpy.util.pack_to_tuple', DeprecationWarning)
    return pack_to_tuple(obj)


def pack_to_list(obj):
    it = type(obj)
    if it == list:
        return obj
    else:
        return [obj]


def list_pack(obj):
    warnings.warn('Use sasoptpy.util.pack_to_list', DeprecationWarning)
    return pack_to_list(obj)


def _wrap_expression_with_iterators(exp, operator, iterators):
    r = exp
    if r.get_name() is None:
        r.set_name(get_next_name())
    if r._operator is None:
        r._operator = operator
    for i in iterators:
        if isinstance(i, sasoptpy.abstract.SetIterator):
            r._iterkey.append(i)
        if isinstance(i, sasoptpy.abstract.SetIteratorGroup):
            r._iterkey.append(i)
    wrapper = sasoptpy.core.Expression()
    wrapper.set_member(key=r.get_name(), ref=r, val=1)
    wrapper._abstract = True
    return wrapper


def expr_nested(argv, operator):
    clocals = argv.gi_frame.f_locals.copy()

    exp = sasoptpy.core.Expression()
    refs = []
    iterators = []
    for i in argv:
        if sasoptpy.core.util.is_expression(i):
            refs.append(i)
            newlocals = argv.gi_frame.f_locals
            for nl in newlocals.keys():
                if nl not in clocals and\
                        (type(newlocals[nl]) == sasoptpy.abstract.SetIterator or
                         type(newlocals[nl]) == sasoptpy.abstract.SetIteratorGroup):
                    iterators.append(newlocals[nl])
                    newlocals[nl].set_name(nl)
    if iterators:
        iterators = sorted(iterators, key=lambda i: i._objorder)
        exp = exp + refs[0]
        exp = _wrap_expression_with_iterators(exp, operator, iterators)
    else:
        exp.set_member(key=get_next_name(), ref=refs, op=', ', val=1)
        exp._operator = 'min'
        exp._abstract = True
        wrapper = sasoptpy.core.Expression()
        wrapper.set_member(key=exp.get_name(), ref=exp, val=1)
        wrapper._abstract = True
        exp = wrapper
    return exp


def get_subindices(keys):
    subindex = []
    for key in keys:
        if isinstance(key, tuple):
            for i in flatten_tuple(key):
                subindex.append(str(i))
        elif isinstance(key, sasoptpy.abstract.SetIteratorGroup):
            subindex.append(key.get_name())
        elif not is_key_abstract(key):
            subindex.append(safe_string(str(key)))
    return subindex


def get_subindices_optmodel_format(subindex):
    if subindex:
        return '_' + '_'.join([str(i) for i in subindex])
    else:
        return ''


def _to_optmodel_loop(keys, parent=None, subindex=True):
    s = ''
    if subindex:
        subindex = get_subindices(keys)
        s += get_subindices_optmodel_format(subindex)
    iters = get_iterators(keys)
    conds = get_conditions(keys)
    if parent is not None:
        conds.extend(get_conditions(parent))
    iter_str = get_iterators_optmodel_format(iters, conds)
    if iter_str != '':
        s += ' ' + iter_str
    return s


def get_iterators(keys):
    """
    Returns a list of definition strings for a given list of SetIterators
    """
    iterators = []
    groups = {}
    for key in keys:
        if is_key_abstract(key):
            iterators.append(key)
    return iterators


def get_iterators_optmodel_format(iters, conds=None):
    if conds is None:
        conds = []

    valid_iterators = []
    for i in iters:
        if sasoptpy.abstract.is_key_abstract(i):
            valid_iterators.append(i)
    s = ''
    if len(valid_iterators) > 0:
        s = '{' + ', '.join([i._get_for_expr() for i in valid_iterators])
        if conds:
            s += ': ' + ' and '.join(conds)
        s += '}'
    return s


def _recursive_walk(obj, func):
    """
    Calls a given method recursively for given objects


    Parameters
    ----------
    func : string
        Name of the method / function be called

    Notes
    -----
    - This function is for internal consumption.

    """
    result = []
    for i in list(obj):
        m_call = getattr(i, func)
        result.append(m_call())
    return result


def is_set_abstract(arg):
    return sasoptpy.abstract.is_abstract_set(arg)


def is_comparable(arg):
    forbidden_types = [pd.Series, pd.DataFrame, dict, tuple]
    if any(isinstance(arg, f) for f in forbidden_types):
        return False
    else:
        return True


def is_model(arg):
    return isinstance(arg, sasoptpy.core.Model)


def is_workspace(arg):
    return isinstance(arg, sasoptpy.session.Workspace)


def get_conditions(keys):
    conditions = []
    for key in keys:
        if sasoptpy.core.util.is_expression(key):
            if key.sym.get_conditions_len() > 0:
                conditions.append(key.sym.get_conditions_str())
    return conditions


def to_condition_expression(exp):
    if hasattr(exp, '_cond_expr'):
        return exp._cond_expr()


def get_mutable(exp):
    """
    Returns a mutable copy of the given expression if it is immutable

    Parameters
    ----------
    exp : :class:`Variable` or :class:`Expression`
        Object to be wrapped

    Returns
    -------
    r : :class:`Expression`
        Mutable copy of the expression, if the original is immutable
    """
    if isinstance(exp, sasoptpy.Variable):
        r = sasoptpy.Expression(exp)
        r._abstract = exp._abstract
    elif isinstance(exp, sasoptpy.SetIterator):
        r = wrap_expression(exp)
        return r
    elif isinstance(exp, sasoptpy.Expression):
        r = exp
    else:
        r = sasoptpy.core.Expression(exp)
    return r


def wrap_expression(e, abstract=False):
    """
    Wraps expression inside another expression
    """
    wrapper = sasoptpy.Expression()

    if hasattr(e, 'get_name') and e.get_name() is not None:
        name = e.get_name()
    else:
        name = get_next_name()

    if sasoptpy.core.util.is_expression(e):
        wrapper.set_member(key=name, ref=e, val=1)
        wrapper._abstract = e._abstract or abstract
    elif isinstance(e, dict):
        wrapper._linCoef[name] = {**e}
    elif isinstance(e, str):
        wrapper = sasoptpy.Auxiliary(base=e)
    elif np.isinstance(type(e), np.number):
        wrapper += e

    return wrapper


def wrap(e, abstract=False):
    warnings.warn('Use sasoptpy.util.wrap_expression', DeprecationWarning)
    return wrap_expression(e, abstract)


def flatten_tuple(tp):
    """
    Flattens nested tuples

    Parameters
    ----------

    tp : tuple
        Nested tuple to be flattened

    Returns
    -------
    elem : Generator-type
        A generator-type object representing the flat tuple

    Examples
    --------
    >>> tp = (3, 4, (5, (1, 0), 2))
    >>> print(list(so.flatten_tuple(tp)))
    [3, 4, 5, 1, 0, 2]

    """
    for elem in tp:
        if isinstance(elem, tuple):
            yield from flatten_tuple(elem)
        else:
            yield elem


def _to_optmodel_quoted_string(item):
    if isinstance(item, int):
        return str(item)
    elif isinstance(item, str):
        return "'{}'".format(item)
    elif isinstance(item, tuple):
        return '<' + ','.join(_to_optmodel_quoted_string(j) for j in item) + '>'
    else:
        return str(item)


def _to_sas_string(obj):
    if hasattr(obj, '_expr'):
        return obj._expr()
    elif isinstance(obj, str):
        return "'{}'".format(obj)
    elif isinstance(obj, tuple):
        return ', '.join(_to_sas_string(i) for i in obj)
    elif isinstance(obj, list):
        return '{{{}}}'.format(','.join([_to_sas_string(i) for i in obj]))
    elif isinstance(obj, range):
        if obj.step == 1:
            return '{}..{}'.format(_to_sas_string(obj.start),
                                   _to_sas_string(obj.stop-1))
        else:
            return '{}..{} by {}'.format(_to_sas_string(obj.start),
                                         _to_sas_string(obj.stop-1),
                                         _to_sas_string(obj.step))
    elif np.isinstance(type(obj), np.number):
        return str(obj)
    elif isinstance(obj, sasoptpy.abstract.Conditional):
        parent = obj._parent
        iters = get_iterators(parent)
        conds = get_conditions(parent)
        return get_iterators_optmodel_format(iters, conds)
    elif obj is None:
        return '.'
    else:
        raise TypeError('Cannot convert type {} to SAS string'.format(type(obj)))


def _insert_brackets(prefix, keys):
    s = prefix
    if len(keys) > 0:
        s += '['
        k = pack_to_tuple(keys)
        s += ', '.join(_to_iterator_expression(k))
        s += ']'
    return s


def _to_iterator_expression(itlist):
    strlist = []
    for i in itlist:
        if isinstance(i, sasoptpy.core.Expression):
            strlist.append(i._expr())
        elif isinstance(i, str):
            strlist.append("'{}'".format(i))
        else:
            strlist.append(str(i))
    return strlist


def get_python_symbol(symbol):
    if symbol == '^':
        return '**'
    else:
        return symbol


def get_unsafe_operators():
    return ['+', '-', '/', '**', '^', '(', ')', 'sum']


def safe_string(st):
    return "".join(c if c.isalnum() else '_' for c in st)


def get_object_order(obj):
    return getattr(obj, '_objorder', None)


def get_attribute_definitions(members):
    definitions = [_attribute_to_defn(i) for i in members]
    defn = '\n'.join(definitions)
    return defn


def _attribute_to_defn(d):
    expr = sasoptpy.util.to_expression
    if d['key'] == 'init':
        return '{} = {};'.format(expr(d['ref']), d['value'])
    if d['key'] == 'lb' or d['key'] == 'ub':
        return '{}.{} = {};'.format(expr(d['ref']), d['key'], d['value'])


def addSpaces(s, space):
    white = ' '*space
    return white + white.join(s.splitlines(1))


def get_session_type(sess):
    if sess is None:
        return None
    else:
        sess_type = type(sess).__name__
        if sess_type == 'CAS':
            return 'CAS'
        elif sess_type == 'SASsession':
            return 'SAS'
        else:
            return None


# def safe_variable_name(name):
#     return name.replace(' ', '_').replace('\'', '')


def get_group_name(name):
    return name.split('[')[0]


def has_expr(e):
    return hasattr(e, '_expr')




#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for core classes.
"""

import unittest
import sasoptpy as so


class TestConstraint(unittest.TestCase):
    """
    Unit tests for :class:`sasoptpy.Constraint` objects
    """

    def setUp(self):
        pass

    def test_constructor(self):
        x = so.Variable(name='x')
        c2 = 3 * x + x ** 2 >= 10
        c3 = so.Constraint(exp=c2, name='c3')
        self.assertEqual(str(c3), "3 * x + (x) ** (2) >=  10")

        def no_direction():
            c4 = so.Constraint(2 * x, name='c4')
        self.assertRaises(AttributeError, no_direction)

        c5 = so.Constraint(c3, name='c5')
        self.assertEqual(str(c5), "3 * x + (x) ** (2) >=  10")

    def test_update_var_coef(self):
        x = so.Variable(name='x')
        y = so.Variable(name='y')
        c1 = so.Constraint(2 * x + 3 * y <= 20, name='c1')

        c1.update_var_coef(x, 5)
        self.assertEqual(str(c1), '5 * x + 3 * y <=  20')

        z = so.Variable(name='z')
        c1.update_var_coef(z, 10)
        self.assertEqual(str(c1), '5 * x + 3 * y + 10 * z <=  20')

    def test_change_direction(self):
        x = so.Variable(name='x')
        y = so.Variable(name='y')
        c1 = so.Constraint(2 * x + 3 * y <= 20, name='c1')

        c1.set_direction('G')
        self.assertEqual(str(c1), '2 * x + 3 * y >=  20')

        def unknown_direction():
            c1.set_direction('X')
        self.assertRaises(ValueError, unknown_direction)

    def test_get_value(self):
        x = so.Variable(name='x')
        y = so.Variable(name='y')
        c1 = so.Constraint(2 * x + 3 * y <= 20, name='c1')

        x.set_value(4)
        y.set_value(3)

        self.assertEqual(c1.get_value(), 17)
        self.assertEqual(c1.get_value(rhs=True), -3)

    def test_definition(self):
        x = so.Variable(name='x')
        c1 = so.Constraint(3 * x == [5, 10], name='c1')
        self.assertEqual(c1._defn(), 'con c1 : 5 <= 3 * x <= 10;')

        c2 = so.Constraint(5 * x == 15, name='c2')
        self.assertEqual(c2._defn(), 'con c2 : 5 * x = 15;')

        c3 = so.Constraint(10 * x >= 5, name='c3')
        self.assertEqual(c3._defn(), 'con c3 : 10 * x >= 5;')

    def test_str(self):
        x = so.Variable(name='x')
        c1 = 2 * x <= 5
        self.assertEqual(str(c1), "2 * x <=  5")

        c2 = 2 * x >= 5
        self.assertEqual(str(c2), "2 * x >=  5")

        c3 = 2 * x == 5
        self.assertEqual(str(c3), "2 * x ==  5")

        c4 = 2 * x == [5, 10]
        self.assertEqual(str(c4), "2 * x ==  [5, 10]")

        def unknown_direction():
            c4._direction = 'X'
            c4s = str(c4)
        self.assertRaises(ValueError, unknown_direction)

    def test_repr(self):
        x = so.Variable(name='x')

        c1 = so.Constraint(2 * x <= 5, name='c1')
        self.assertEqual(repr(c1), "sasoptpy.Constraint(2 * x <=  5, name='c1')")

        c2 = 2 * x <= 5
        self.assertEqual(repr(c2), "sasoptpy.Constraint(2 * x <=  5, name=None)")

    def tearDown(self):
        so.reset()


import sasoptpy
from sasoptpy.util import to_expression
from .set import Set


class AbstractRange(Set):

    def __init__(self, start, end, step=1):
        super().__init__(name=None)
        self._start = start
        self._end = end
        self._step = step

    def _expr(self):
        if self._step == 1:
            return '{}..{}'.format(to_expression(self._start).replace(' ', ''),
                                   to_expression(self._end).replace(' ', ''))
        else:
            return '{}..{} by {}'.format(
                to_expression(self._start),
                to_expression(self._end),
                to_expression(self._step)
            )

    def _defn(self):
        pass

    def __str__(self):
        return self._expr()



import sasoptpy as so
import pandas as pd


def test(cas_conn):

    m = so.Model(name='farm_planning', session=cas_conn)

    # Input Data

    cow_data_raw = []
    for age in range(12):
        if age < 2:
            row = {'age': age,
                   'init_num_cows': 10,
                   'acres_needed': 2/3.0,
                   'annual_loss': 0.05,
                   'bullock_yield': 0,
                   'heifer_yield': 0,
                   'milk_revenue': 0,
                   'grain_req': 0,
                   'sugar_beet_req': 0,
                   'labour_req': 10,
                   'other_costs': 50}
        else:
            row = {'age': age,
                   'init_num_cows': 10,
                   'acres_needed': 1,
                   'annual_loss': 0.02,
                   'bullock_yield': 1.1/2,
                   'heifer_yield': 1.1/2,
                   'milk_revenue': 370,
                   'grain_req': 0.6,
                   'sugar_beet_req': 0.7,
                   'labour_req': 42,
                   'other_costs': 100}
        cow_data_raw.append(row)
    cow_data = pd.DataFrame(cow_data_raw).set_index(['age'])
    grain_data = pd.DataFrame([
        ['group1', 20, 1.1],
        ['group2', 30, 0.9],
        ['group3', 20, 0.8],
        ['group4', 10, 0.65]
        ], columns=['group', 'acres', 'yield']).set_index(['group'])
    num_years = 5
    num_acres = 200
    bullock_revenue = 30
    heifer_revenue = 40
    dairy_cow_selling_age = 12
    dairy_cow_selling_revenue = 120
    max_num_cows = 130
    sugar_beet_yield = 1.5
    grain_cost = 90
    grain_revenue = 75
    grain_labour_req = 4
    grain_other_costs = 15
    sugar_beet_cost = 70
    sugar_beet_revenue = 58
    sugar_beet_labour_req = 14
    sugar_beet_other_costs = 10
    nominal_labour_cost = 4000
    nominal_labour_hours = 5500
    excess_labour_cost = 1.2
    capital_outlay_unit = 200
    num_loan_years = 10
    annual_interest_rate = 0.15
    max_decrease_ratio = 0.50
    max_increase_ratio = 0.75

    # Sets

    AGES = cow_data.index.tolist()
    init_num_cows = cow_data['init_num_cows']
    acres_needed = cow_data['acres_needed']
    annual_loss = cow_data['annual_loss']
    bullock_yield = cow_data['bullock_yield']
    heifer_yield = cow_data['heifer_yield']
    milk_revenue = cow_data['milk_revenue']
    grain_req = cow_data['grain_req']
    sugar_beet_req = cow_data['sugar_beet_req']
    cow_labour_req = cow_data['labour_req']
    cow_other_costs = cow_data['other_costs']

    YEARS = list(range(1, num_years+1))
    YEARS0 = [0] + YEARS

    # Variables

    numCows = m.add_variables(AGES + [dairy_cow_selling_age], YEARS0, lb=0,
                              name='numCows')
    for age in AGES:
        numCows[age, 0].set_bounds(lb=init_num_cows[age],
                                   ub=init_num_cows[age])
    numCows[dairy_cow_selling_age, 0].set_bounds(lb=0, ub=0)

    numBullocksSold = m.add_variables(YEARS, lb=0, name='numBullocksSold')
    numHeifersSold = m.add_variables(YEARS, lb=0, name='numHeifersSold')

    GROUPS = grain_data.index.tolist()
    acres = grain_data['acres']
    grain_yield = grain_data['yield']
    grainAcres = m.add_variables(GROUPS, YEARS, lb=0, name='grainAcres')
    for group in GROUPS:
        for year in YEARS:
            grainAcres[group, year].set_bounds(ub=acres[group])
    grainBought = m.add_variables(YEARS, lb=0, name='grainBought')
    grainSold = m.add_variables(YEARS, lb=0, name='grainSold')

    sugarBeetAcres = m.add_variables(YEARS, lb=0, name='sugarBeetAcres')
    sugarBeetBought = m.add_variables(YEARS, lb=0, name='sugarBeetBought')
    sugarBeetSold = m.add_variables(YEARS, lb=0, name='sugarBeetSold')

    numExcessLabourHours = m.add_variables(YEARS, lb=0,
                                           name='numExcessLabourHours')
    capitalOutlay = m.add_variables(YEARS, lb=0, name='capitalOutlay')

    yearly_loan_payment = (annual_interest_rate * capital_outlay_unit) /\
                          (1 - (1+annual_interest_rate)**(-num_loan_years))

    # Objective function

    revenue = {year:
               bullock_revenue * numBullocksSold[year] +
               heifer_revenue * numHeifersSold[year] +
               dairy_cow_selling_revenue * numCows[dairy_cow_selling_age,
                                                   year] +
               so.expr_sum(milk_revenue[age] * numCows[age, year]
                            for age in AGES) +
               grain_revenue * grainSold[year] +
               sugar_beet_revenue * sugarBeetSold[year]
               for year in YEARS}

    cost = {year:
            grain_cost * grainBought[year] +
            sugar_beet_cost * sugarBeetBought[year] +
            nominal_labour_cost +
            excess_labour_cost * numExcessLabourHours[year] +
            so.expr_sum(cow_other_costs[age] * numCows[age, year]
                         for age in AGES) +
            so.expr_sum(grain_other_costs * grainAcres[group, year]
                         for group in GROUPS) +
            sugar_beet_other_costs * sugarBeetAcres[year] +
            so.expr_sum(yearly_loan_payment * capitalOutlay[y]
                         for y in YEARS if y <= year)
            for year in YEARS}
    profit = {year: revenue[year] - cost[year] for year in YEARS}

    totalProfit = so.expr_sum(profit[year] -
                               yearly_loan_payment * (num_years - 1 + year) *
                               capitalOutlay[year] for year in YEARS)

    m.set_objective(totalProfit, sense=so.MAX, name='totalProfit')

    # Constraints

    m.add_constraints((
        so.expr_sum(acres_needed[age] * numCows[age, year] for age in AGES) +
        so.expr_sum(grainAcres[group, year] for group in GROUPS) +
        sugarBeetAcres[year] <= num_acres
        for year in YEARS), name='num_acres')

    m.add_constraints((
        numCows[age+1, year+1] == (1-annual_loss[age]) * numCows[age, year]
        for age in AGES if age != dairy_cow_selling_age
        for year in YEARS0 if year != num_years), name='aging')

    m.add_constraints((
        numBullocksSold[year] == so.expr_sum(
            bullock_yield[age] * numCows[age, year] for age in AGES)
        for year in YEARS), name='numBullocksSold_def')

    m.add_constraints((
        numCows[0, year] == so.expr_sum(
            heifer_yield[age] * numCows[age, year]
            for age in AGES) - numHeifersSold[year]
        for year in YEARS), name='numHeifersSold_def')

    m.add_constraints((
        so.expr_sum(numCows[age, year] for age in AGES) <= max_num_cows +
        so.expr_sum(capitalOutlay[y] for y in YEARS if y <= year)
        for year in YEARS), name='max_num_cows_def')

    grainGrown = {(group, year): grain_yield[group] * grainAcres[group, year]
                  for group in GROUPS for year in YEARS}
    m.add_constraints((
        so.expr_sum(grain_req[age] * numCows[age, year] for age in AGES) <=
        so.expr_sum(grainGrown[group, year] for group in GROUPS)
        + grainBought[year] - grainSold[year]
        for year in YEARS), name='grain_req_def')

    sugarBeetGrown = {(year): sugar_beet_yield * sugarBeetAcres[year]
                      for year in YEARS}
    m.add_constraints((
        so.expr_sum(sugar_beet_req[age] * numCows[age, year] for age in AGES)
        <=
        sugarBeetGrown[year] + sugarBeetBought[year] - sugarBeetSold[year]
        for year in YEARS), name='sugar_beet_req_def')

    m.add_constraints((
        so.expr_sum(cow_labour_req[age] * numCows[age, year]
                     for age in AGES) +
        so.expr_sum(grain_labour_req * grainAcres[group, year]
                     for group in GROUPS) +
        sugar_beet_labour_req * sugarBeetAcres[year] <=
        nominal_labour_hours + numExcessLabourHours[year]
        for year in YEARS), name='labour_req_def')
    m.add_constraints((profit[year] >= 0 for year in YEARS), name='cash_flow')

    m.add_constraint(so.expr_sum(numCows[age, num_years] for age in AGES
                                  if age >= 2) /
                     sum(init_num_cows[age] for age in AGES if age >= 2) ==
                     [1-max_decrease_ratio, 1+max_increase_ratio],
                     name='final_dairy_cows_range')

    res = m.solve()

    if res is not None:
        so.pd.display_all()
        print(so.get_solution_table(numCows))
        revenue_df = so.dict_to_frame(revenue, cols=['revenue'])
        cost_df = so.dict_to_frame(cost, cols=['cost'])
        profit_df = so.dict_to_frame(profit, cols=['profit'])
        print(so.get_solution_table(numBullocksSold, numHeifersSold,
                                    capitalOutlay, numExcessLabourHours,
                                    revenue_df, cost_df, profit_df))
        gg_df = so.dict_to_frame(grainGrown, cols=['grainGrown'])
        print(so.get_solution_table(grainAcres, gg_df))
        sbg_df = so.dict_to_frame(sugarBeetGrown, cols=['sugerBeetGrown'])
        print(so.get_solution_table(
            grainBought, grainSold, sugarBeetAcres,
            sbg_df, sugarBeetBought, sugarBeetSold))
        num_acres = m.get_constraint('num_acres')
        na_df = num_acres.get_expressions()
        max_num_cows_con = m.get_constraint('max_num_cows_def')
        mnc_df = max_num_cows_con.get_expressions()
        print(so.get_solution_table(na_df, mnc_df))

    return m.get_objective_value()



from .statement_base import Statement

import sasoptpy


class ObjectiveStatement(Statement):

    def __init__(self, expression, **kwargs):
        super().__init__()
        self.model = kwargs.get('model', None)
        self.name = kwargs.get('name')
        self.expr = expression
        self.sense = kwargs.get('sense')

    def append(self):
        pass

    def _defn(self):
        return '{} {} = {};'.format(self.sense, self.name, self.expr._expr())

    @classmethod
    def set_objective(cls, expression, name, sense):
        st = ObjectiveStatement(expression, name=name, sense=sense)
        return st



from .for_loop import ForLoopStatement
import sasoptpy


class CoForLoopStatement(ForLoopStatement):

    def __init__(self, *args):
        super().__init__(*args)
        self.keyword = 'cofor'

    @classmethod
    def cofor_loop(cls, *args):
        loop = CoForLoopStatement(*args)
        return loop



from .statement_base import Statement

import sasoptpy


class PrintStatement(Statement):

    def __init__(self, *args):
        super().__init__()
        for arg in args:
            self.elements.append(arg)
        self._print_type = 'print'
        self._print_names = False

    def append(self, arg):
        self.elements.append(arg)

    def set_print_type(self, print_type):
        self._print_type = print_type

    def allow_print_names(self):
        self._print_names = True

    def _defn(self):
        keyword = f'{self._print_type}'
        items = []
        for i in self.elements:
            word = sasoptpy.to_expression(i)
            if self._print_names:
                word += '='
            items.append(word)
        return keyword + ' ' + ' '.join(items) + ';'

    @classmethod
    def print_item(cls, *args):
        ps = PrintStatement(*args)
        return ps

    @classmethod
    def put_item(cls, *args, names=False):
        ps = PrintStatement(*args)
        ps.set_print_type('put')
        if names:
            ps.allow_print_names()
        return ps



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
User accessible utility functions
"""

from collections import Iterable
import random
import string
import warnings
from contextlib import contextmanager

import sasoptpy
from sasoptpy.libs import (pd, np)
from .package_utils import (
    wrap_expression, _wrap_expression_with_iterators,
    get_first_member, pack_to_tuple, expr_nested)


def concat(exp1, exp2):
    return wrap_expression(exp1).concat(wrap_expression(exp2))


@contextmanager
def iterate(set, name):
    if not sasoptpy.abstract.is_abstract_set(set):
        set = sasoptpy.Set.from_object(set)
    if isinstance(name, list):
        yield sasoptpy.abstract.SetIteratorGroup(set, names=name)
    else:
        yield sasoptpy.abstract.SetIterator(set, name=name)


def exp_range(start, stop, step=1):
    """
    Creates a set within specified range

    Parameters
    ----------
    start : :class:`Expression`
        First value of the range
    stop : :class:`Expression`
        Last value of the range
    step : :class:`Expression`, optional
        Step size of the range

    Returns
    -------
    exset : Set
        Set that represents the range

    Examples
    --------

    >>> N = so.Parameter(name='N')
    >>> p = so.exp_range(1, N)
    >>> print(p._defn())
    set 1..N;

    """
    regular = isinstance(start, int) and isinstance(stop, int) and\
        isinstance(step, int)
    if regular:
        return range(start, stop, step)
    s = sasoptpy.abstract.AbstractRange(start, stop, step)
    return s


def get_solution_table(*argv, key=None, rhs=False):
    return get_value_table(*argv, key=key, rhs=rhs)


def get_value_table(*args, **kwargs):
    """
    Returns values of the given arguments as a merged pandas DataFrame

    Parameters
    ----------
    key : list, optional
        Keys for objects
    rhs : bool, optional
        Option for including constant values

    Returns
    -------
    table : :class:`pandas.DataFrame`
        DataFrame object that holds object values
    """

    if len(args) == 0:
        return None

    series = []
    for arg in args:
        s = get_values(arg, **kwargs)
        series.append(s)
    return pd.concat(series, axis=1, sort=False)


def get_values(arg, **kwargs):
    """
    Returns values of given set of arguments as a pandas Series
    """
    if isinstance(arg, pd.Series):
        arg_values = arg.apply(
            lambda row: row.get_value() if hasattr(row, 'get_value')
            else str(row))
        return arg_values
    elif isinstance(arg, pd.DataFrame):
        arg_values = arg.apply(
            lambda col: col.apply(
                lambda row: row.get_value() if hasattr(row, 'get_value')
                else str(row)))
        return arg_values
    elif isinstance(arg, sasoptpy.VariableGroup):
        keys = []
        values = []
        members = arg.get_members() if arg._abstract is False else arg.get_shadow_members()
        for i in members:
            if any([sasoptpy.abstract.is_abstract(j) for j in i]):
            #if sasoptpy.abstract.is_abstract(get_first_member(i)):
                continue
            keys.append(get_first_member(i))
            values.append(members[i].get_value())
        return pd.Series(values, index=keys, name=arg.get_name())
    elif isinstance(arg, sasoptpy.ConstraintGroup):
        keys = []
        values = []
        members = arg.get_all_keys()
        rhs = kwargs.get('rhs', False)
        for i in members:
            if sasoptpy.abstract.is_abstract(get_first_member(i)):
                continue
            keys.append(get_first_member(i))
            values.append(arg[i].get_value(rhs=rhs))
        return pd.Series(values, index=keys, name=arg.get_name())
    elif isinstance(arg, sasoptpy.ImplicitVar):
        keys = []
        values = []
        members = arg.get_dict()
        for i in members:
            keys.append(get_first_member(i))
            values.append(members[i].get_value())
        return pd.Series(values, index=keys, name=arg.get_name())
    elif isinstance(arg, sasoptpy.Expression):
        return pd.Series([arg.get_value()], index=['-'], name=arg.get_name())
    else:
        return pd.Series([arg], index=['-'], name=str(arg))


def submit(caller, *args, **kwargs):
    session = caller.get_session()
    session_type = sasoptpy.util.package_utils.get_session_type(session)
    if session is None or session_type is None:
        raise RuntimeError('No session is available')

    mediator_class = sasoptpy.mediators[session_type]
    mediator = mediator_class(caller, session)
    return mediator.submit(*args, **kwargs)


def submit_for_solve(caller, *args, **kwargs):
    session = caller.get_session()
    session_type = sasoptpy.util.package_utils.get_session_type(session)
    if session is None or session_type is None:
        raise RuntimeError('No session is available')

    mediator_class = sasoptpy.mediators[session_type]
    mediator = mediator_class(caller, session)
    return mediator.solve(*args, **kwargs)


def submit_for_tune(caller, **kwargs):
    session = caller.get_session()
    session_type = sasoptpy.util.package_utils.get_session_type(session)
    if session_type != 'CAS':
        raise TypeError('Tune action is only available on CAS (SAS Viya) servers.')

    mediator_class= sasoptpy.mediators[session_type]
    mediator = mediator_class(caller, session)
    return mediator.tune(**kwargs)


def quick_sum(argv):
    """
    Summation function for :class:`Expression` objects

    Notes
    -----

    This method will deprecate in future versions.
    Use :func:`expr_sum` instead.

    """
    return sasoptpy.util.expr_sum(argv)


def expr_sum(argv):
    """
    Summation function for :class:`Expression` objects

    Returns
    -------
    exp : :class:`Expression`
        Sum of given arguments

    Examples
    --------

    >>> x = so.VariableGroup(10000, name='x')
    >>> y = so.expr_sum(2*x[i] for i in range(10000))

    Notes
    -----

    This function is faster for expressions compared to Python's native sum()
    function.

    """
    clocals = argv.gi_frame.f_locals.copy()

    exp = sasoptpy.core.Expression()
    exp.set_temporary()
    iterators = []
    for i in argv:
        exp = exp + i
        if sasoptpy.core.util.is_expression(i):
            #if i._abstract:
            newlocals = argv.gi_frame.f_locals
            for nl in newlocals.keys():
                if nl not in clocals and\
                        (type(newlocals[nl]) == sasoptpy.abstract.SetIterator or
                         type(newlocals[nl]) == sasoptpy.abstract.SetIteratorGroup):
                    iterators.append(newlocals[nl])
                    newlocals[nl].set_name(nl)
    if iterators:
        iterators = sorted(iterators, key=lambda i: i._objorder)
        exp = _wrap_expression_with_iterators(exp, 'sum', iterators)
    exp.set_permanent()
    return exp


def expr_min(argv):
    """
    Min aggregator function for :class:`Expression` objects

    Returns
    -------
    exp : :class:`Expression`
        Min of given arguments

    Examples
    --------

    >>> x = so.VariableGroup(10, name='x')
    >>> y = so.expr_min(2*x[i] for i in range(10))

    """
    return expr_nested(argv, 'min')


def expr_max(argv):
    """
    Max aggregator function for :class:`Expression` objects

    Returns
    -------
    exp : :class:`Expression`
        Max of given arguments

    Examples
    --------

    >>> x = so.VariableGroup(10, name='x')
    >>> y = so.expr_max(2*x[i] for i in range(10))

    """
    return expr_nested(argv, 'max')


def reset():
    """
    Resets package configs and internal counters
    """
    sasoptpy.config.reset()
    sasoptpy.itemid = 0


def reset_globals():
    warnings.warn('Use sasoptpy.reset()', DeprecationWarning)
    reset()


def dict_to_frame(dictobj, cols=None):
    """
    Converts dictionaries to DataFrame objects for pretty printing

    Parameters
    ----------
    dictobj : dict
        Dictionary to be converted
    cols : list, optional
        Column names

    Returns
    -------
    frobj : DataFrame
        DataFrame representation of the dictionary

    Examples
    --------

    >>> d = {'coal': {'period1': 1, 'period2': 5, 'period3': 7},
    >>>      'steel': {'period1': 8, 'period2': 4, 'period3': 3},
    >>>      'copper': {'period1': 5, 'period2': 7, 'period3': 9}}
    >>> df = so.dict_to_frame(d)
    >>> print(df)
            period1  period2  period3
    coal          1        5        7
    copper        5        7        9
    steel         8        4        3

    """
    frobj = pd.convert_dict_to_frame(dictobj, cols)
    return frobj


def flatten_frame(df, swap=False):
    """
    Converts a :class:`pandas.DataFrame` object into :class:`pandas.Series`

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        DataFrame to be flattened
    swap : boolean, optional
        Option to use columns as first index

    Returns
    -------
    new_frame : :class:`pandas.DataFrame`
        A new DataFrame where indices consist of index and columns names as
        tuples

    Examples
    --------

    >>> price = pd.DataFrame([
    >>>     [1, 5, 7],
    >>>     [8, 4, 3],
    >>>     [5, 7, 9]], columns=[\'period1\', \'period2\', \'period3\']).\\
    >>>     set_index([[\'coal\', \'steel\', \'copper\']])
    >>> print(\'Price data: \\n{}\'.format(price))
    >>> price_f = so.flatten_frame(price)
    >>> print(\'Price data: \\n{}\'.format(price_f))
    Price data:
            period1  period2  period3
    coal          1        5        7
    steel         8        4        3
    copper        5        7        9
    Price data:
    (coal, period1)      1
    (coal, period2)      5
    (coal, period3)      7
    (steel, period1)     8
    (steel, period2)     4
    (steel, period3)     3
    (copper, period1)    5
    (copper, period2)    7
    (copper, period3)    9
    dtype: int64

    """
    new_frame = df.stack()
    if swap:
        new_frame = new_frame.swaplevel()
    #new_frame.index = new_frame.index.to_series()
    return new_frame


def is_linear(item):
    return item._is_linear()


def has_integer_variables(item):
    return item._has_integer_vars()


def export_to_mps(model, filename=None):
    mps = model.to_mps()
    # Convert to standard MPS format
    mps._set_value(0, 'Field2', mps.loc[0, 'Field3'])
    mps._set_value(0, 'Field3', '')
    mps._set_value(0, 'Field4', np.nan)
    mps._set_value(0, 'Field6', np.nan)
    mps._set_value(len(mps)-1, 'Field4', np.nan)
    mps._set_value(len(mps)-1, 'Field6', np.nan)
    mps.drop(columns="_id_", inplace=True)
    def add_space(r):
        keywords = ['NAME', 'ROWS', 'COLUMNS', 'RHS', 'BOUNDS', 'RANGES', 'ENDATA']
        if r not in keywords:
            return " " + r
        else:
            return r
    mps['Field1'] = mps['Field1'].apply(add_space)

    formatters={
        'Field1': '{{:<{}s}}'.format(mps['Field1'].str.len().max()).format,
        'Field2': '{{:<{}s}}'.format(mps['Field2'].str.len().max()).format,
        'Field3': '{{:<{}s}}'.format(mps['Field3'].str.len().max()).format,
        'Field5': '{{:<{}s}}'.format(mps['Field5'].str.len().max()).format,
        }

    mps_str = mps.to_string(formatters=formatters, index=False, header=False, na_rep='')
    if mps_str[0] == " ":
        mps_str = '\n'.join([i[1:] for i in mps_str.split('\n')])

    if filename is not None:
        with open(filename, 'w') as file:
            file.write(mps_str)

    return mps_str




from math import inf
import warnings

import sasoptpy
from sasoptpy.libs import np, pd


def is_expression(obj):
    return isinstance(obj, sasoptpy.core.Expression)


def is_variable(obj):
    return isinstance(obj, sasoptpy.core.Variable)


def is_constraint(obj):
    return isinstance(obj, sasoptpy.core.Constraint)


def is_droppable(obj):
    return isinstance(obj, sasoptpy.core.Constraint) or \
           isinstance(obj, sasoptpy.core.ConstraintGroup)


def is_model(obj):
    return isinstance(obj, sasoptpy.core.Model)


def is_abstract(obj):
    if hasattr(obj, '_abstract') and obj._abstract:
        return True
    return False


def _to_iterator_expression(itlist):
    strlist = []
    for i in itlist:
        if isinstance(i, sasoptpy.core.Expression):
            strlist.append(i._expr())
        elif isinstance(i, str):
            strlist.append("'{}'".format(i))
        else:
            strlist.append(str(i))
    return strlist


def _to_safe_iterator_expression(itlist):
    strlist = []
    for i in itlist:
        if isinstance(i, sasoptpy.core.Expression):
            strlist.append(i._expr())
        elif isinstance(i, str):
            safe_i = "".join(c if c.isalnum() else '_' for c in i)
            strlist.append("'{}'".format(safe_i))
        else:
            strlist.append(str(i))
    return strlist


def is_key_empty(iterkey):
    return str(iterkey) == "('',)"


def _evaluate(comp):
    """
    Evaluates the value of a given expression component.

    Parameters
    ----------
    comp : dict
        Dictionary of references, coefficient and operator

    Returns
    -------
    v : float
        Current value of the expression.
    """

    ref = comp['ref']
    val = comp['val']
    op = comp.get('op')
    v = None

    if op is None:
        op = '*'

    if op == '*':
        v = val
        for i in ref:
            v = v * i.get_value()
    elif op == '/':
        try:
            v = val * ref[0].get_value() / ref[1].get_value()
        except ZeroDivisionError:
            raise ZeroDivisionError('Division error in evaluation')
    elif op == '^':
        v = val * ref[0].get_value() ** ref[1].get_value()

    return v


def expression_to_constraint(left, relation, right):

    if isinstance(right, list) and relation == 'E':
        e = left.copy()
        e.add_to_member_value('CONST', -1 * min(right[0], right[1]))
        ranged_constraint = sasoptpy.core.Constraint(
            exp=e, direction='E', crange=abs(right[1] - right[0]))
        return ranged_constraint
    elif not is_variable(left):
        if left._operator is None:
            r = left.copy()
        else:
            r = sasoptpy.core.Expression(0)
            r += left
        if np.isinstance(type(right), np.number):
            r._linCoef['CONST']['val'] -= right
            right = 0
        elif is_expression(right):
            r -= right
        elif right is None:
            r.set_member_value('CONST', np.nan)
        generated_constraint = sasoptpy.core.Constraint(
            exp=r, direction=relation, crange=0, internal=True)
        return generated_constraint
    else:
        r = sasoptpy.core.Expression()
        for v in left._linCoef:
            r._add_coef_value(left._linCoef[v]['ref'], v,
                              left._linCoef[v]['val'])
        if np.isinstance(type(right), np.number):
            r._linCoef['CONST']['val'] -= right
        else:
            for v in right._linCoef:
                if r.get_member(v):
                    r.add_to_member_value(v, -1 * right.get_member_value(v))
                else:
                    r.copy_member(v, right)
                    r.mult_member_value(v, -1)
        generated_constraint = sasoptpy.core.Constraint(
            exp=r, direction=relation, crange=0)
        return generated_constraint


def get_key_for_expr(key):
    if isinstance(key, sasoptpy.abstract.SetIterator):
        return 'for ' + key._get_for_expr()
    else:
        return 'for {} in {}'.format(key.get_name(), key._set.get_name())


def multiply_coefficients(left, right, target):

    left = left.copy()
    left['CONST'] = left.pop('CONST')
    right = right.copy()
    right['CONST'] = right.pop('CONST')

    for i, x in left.items():
        for j, y in right.items():
            if x['ref'] is None and y['ref'] is None:
                append_coef_value(
                    target, ref=None, key='CONST', value=x['val'] * y['val'])
            elif x['ref'] is None:
                if x['val'] * y['val'] != 0:
                    append_coef_value(
                        target, ref=y['ref'], key=j, value=x['val'] * y['val'])
            elif y['ref'] is None:
                if x['val'] * y['val'] != 0:
                    append_coef_value(
                        target, ref=x['ref'], key=i, value=x['val'] * y['val'])
            else:
                if x['val'] * y['val'] != 0:
                    if x.get('op') is None and y.get('op') is None:
                        new_key = sasoptpy.util.pack_to_tuple(i) + \
                                  sasoptpy.util.pack_to_tuple(j)
                        alt_key = sasoptpy.util.pack_to_tuple(j) + \
                                  sasoptpy.util.pack_to_tuple(i)
                        new_ref = list(x['ref']) + list(y['ref'])
                        if alt_key in target:
                            append_coef_value(
                                target, ref=new_ref, key=alt_key,
                                value=x['val'] * y['val'])
                        else:
                            append_coef_value(
                                target, ref=new_ref, key=new_key,
                                value=x['val'] * y['val'])
                    else:
                        new_key = (i, j)
                        x_actual = x['ref']
                        if 'op' in x and x['op'] is not None:
                            x_actual = sasoptpy.util.wrap_expression(x)
                        y_actual = y['ref']
                        if 'op' in y and y['op'] is not None:
                            y_actual = sasoptpy.util.wrap_expression(y)

                        append_coef_value(
                            target, ref=[x_actual, y_actual], key=new_key,
                            value=x['val'] * y['val'])


def append_coef_value(original_dict, ref, key, value):
    if key in original_dict:
        original_dict[key]['val'] += value
    else:
        original_dict[key] = {'ref': ref, 'val': value}


def get_name_from_keys(name, keys):
    return '{}['.format(name) + ','.join(format(k) for k in keys) + ']'


def get_generator_names(argv):
    keys = tuple()
    vnames = argv.gi_code.co_cellvars + argv.gi_code.co_varnames
    vdict = argv.gi_frame.f_locals
    for ky in vnames:
        if ky != '.0':
            keys = keys + (vdict[ky],)
    return keys


def get_default_value(vartype, key):
    return sasoptpy.config['default_bounds'].get(vartype).get(key)


def get_default_bounds_if_none(vartype, lb, ub):
    lb = get_default_value(vartype, 'lb') if lb is None else lb
    ub = get_default_value(vartype, 'ub') if ub is None else ub
    return lb, ub


def is_regular_component(cm):
    if not hasattr(cm, '_objorder'):
        return False
    if cm._objorder > 0:
        if hasattr(cm, '_shadow') and cm._shadow is True:
            return False
        if hasattr(cm, '_parent') and cm._parent is not None:
            return False
    return True


def has_parent(cm):
    return hasattr(cm, '_parent') and cm._parent is not None


def get_group_bound(bound):
    group_types = [pd.DataFrame, pd.Series, list, dict, tuple]
    if any(isinstance(bound, g) for g in group_types):
        return None
    else:
        return bound


def is_valid_lb(lb, variable_type):
    if lb is None:
        return False
    if lb == -inf or lb == -np.inf:
        return False
    if lb == 0 and variable_type == sasoptpy.BIN:
        return False

    if np.isinstance(type(lb), np.number):
        return True
    else:
        return False


def is_valid_ub(ub, variable_type):
    if ub is None:
        return False
    if ub == inf or ub == np.inf:
        return False
    if ub == 1 and variable_type == sasoptpy.BIN:
        return False

    if np.isinstance(type(ub), np.number):
        return True
    else:
        return True


def is_valid_init(init, varible_type):
    if init is None:
        return False
    if np.isinstance(type(init), np.number):
        return True
    return False




from .statement_base import Statement
import sasoptpy


class FixStatement(Statement):

    def __init__(self, *elements):
        super().__init__()
        self.keyword = 'fix'
        for i in elements:
            self.append(i)

    def append(self, element):
        self.elements.append(element)

    def _defn(self):
        elems = []
        for i in self.elements:
            elems.append('{}={}'.format(
                sasoptpy.to_expression(i[0]),
                sasoptpy.to_expression(i[1])
            ))
        s = self.keyword + ' ' + ' '.join(elems) + ';'
        return s

    @classmethod
    def fix(cls, *items):
        if len(items) == 2 and not any(isinstance(i, tuple) for i in items):
            items = ((items[0], items[1]),)
        fs = FixStatement(*items)
        return fs


class UnfixStatement(Statement):

    def __init__(self, *elements):
        super().__init__()
        self.keyword = 'unfix'
        for i in elements:
            self.append(i)

    def append(self, element):
        self.elements.append(element)

    def _defn(self):
        elems = []
        for i in self.elements:
            if isinstance(i, tuple):
                elems.append('{}={}'.format(
                    sasoptpy.to_expression(i[0]),
                    sasoptpy.to_expression(i[1])
                ))
            else:
                elems.append(sasoptpy.to_expression(i))
        s = self.keyword + ' ' + ' '.join(elems) + ';'
        return s

    @classmethod
    def unfix(cls, *items):
        fs = UnfixStatement(*items)
        return fs



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Math includes the definition of mathematical operations

"""

import math
import sasoptpy.util


func_equivalent = {
    'abs': abs,
    'log': math.log,
    'log2': math.log2,
    'log10': math.log10,
    'exp': math.exp,
    'sqrt': math.sqrt,
    'mod': lambda x,y: divmod(x,y)[1],
    'int': int,
    'sign': lambda x: math.copysign(1, x),
    'max': max,
    'min': min,
    'sin': math.sin,
    'cos': math.cos,
    'tan': math.tan,
    'sinh': math.sinh,
    'cosh': math.cosh,
    'tanh': math.tanh
}


def math_func(exp, op, *args):
    """
    Function wrapper for math functions

    Parameters
    ----------
    exp : :class:`Expression`
        Expression where the math function will be applied
    op : string
        String representation of the math function
    args : float, optional
        Additional arguments
    """
    exp = sasoptpy.util.get_mutable(exp)
    exp._operator = op
    for arg in args:
        exp._arguments.append(arg)
    r = sasoptpy.util.wrap_expression(exp)
    return r


# Basic functions
def abs(exp):
    """
    Absolute value function
    """
    return math_func(exp, 'abs')


def log(exp):
    """
    Natural logarithm function
    """
    return math_func(exp, 'log')


def log2(exp):
    """
    Logarithm function in base 2
    """
    return math_func(exp, 'log2')


def log10(exp):
    """
    Logarithm function in base 10
    """
    return math_func(exp, 'log10')


def exp(exp):
    """
    Exponential function
    """
    return math_func(exp, 'exp')


def sqrt(exp):
    """
    Square root function
    """
    return math_func(exp, 'sqrt')


def mod(exp, divisor):
    """
    Modulo function

    Parameters
    ----------
    exp : :class:`Expression`
        Dividend
    divisor : :class:`Expression`
        Divisor
    """
    return math_func(exp, 'mod', divisor)


def int(exp):
    """
    Integer value function
    """
    return math_func(exp, 'int')


def sign(exp):
    """
    Sign value function
    """
    return math_func(exp, 'sign')


# Descriptive functions
def max(exp, *args):
    """
    Largest value function
    """
    return math_func(exp, 'max', *args)


def min(exp, *args):
    """
    Smallest value function
    """
    return math_func(exp, 'min', *args)


# Trigonometric functions
def sin(exp):
    """
    Sine function
    """
    return math_func(exp, 'sin')


def cos(exp):
    """
    Cosine function
    """
    return math_func(exp, 'cos')


def tan(exp):
    """
    Tangent function
    """
    return math_func(exp, 'tan')


def sinh(exp):
    """
    Hyperbolic sine function
    """
    return math_func(exp, 'sinh')


def cosh(exp):
    """
    Hyperbolic cosine function
    """
    return math_func(exp, 'cosh')


def tanh(exp):
    """
    Hyperbolic tangent function
    """
    return math_func(exp, 'tanh')




import sasoptpy as so


def test(cas_conn):

    m = so.Model(name='nlpse01', session=cas_conn)
    x = m.add_variables(range(1, 9), lb=0.1, ub=10, name='x')

    f = so.Expression(0.4 * (x[1]/x[7]) ** 0.67 + 0.4 * (x[2]/x[8]) ** 0.67 + 10 - x[1] - x[2], name='f')
    m.set_objective(f, sense=so.MIN, name='f1')

    m.add_constraint(1 - 0.0588*x[5]*x[7] - 0.1*x[1] >= 0, name='c1')
    m.add_constraint(1 - 0.0588*x[6]*x[8] - 0.1*x[1] - 0.1*x[2] >= 0, name='c2')
    m.add_constraint(1 - 4*x[3]/x[5] - 2/(x[3]**0.71 * x[5]) - 0.0588*(x[7]/x[3]**1.3) >= 0, name='c3')
    m.add_constraint(1 - 4*x[4]/x[6] - 2/(x[4]**0.71 * x[6]) - 0.0588*(x[8]/x[4]**1.3) >= 0, name='c4')
    m.add_constraint(f == [0.1, 4.2], name='frange')

    x[1].set_init(6)
    x[2].set_init(3)
    x[3].set_init(0.4)
    x[4].set_init(0.2)
    x[5].set_init(6)
    x[6].set_init(6)
    x[7].set_init(1)
    x[8].set_init(0.5)

    m.solve(verbose=True, options={'with': 'nlp', 'algorithm': 'activeset'})

    print(m.get_problem_summary())
    print(m.get_solution_summary())
    if m.get_session_type() == 'CAS':
        print(m.get_solution()[['var', 'value']])

    return m.get_objective_value()




#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from abc import ABC, abstractmethod
from collections import OrderedDict

import sasoptpy


class Statement(ABC):
    """
    Creates a statement to be executed at the server

    This class is an abstract base class for all statement types.
    """

    def __init__(self):
        import sasoptpy.util
        self.parent = None
        self.header = OrderedDict()
        self.elements = list()
        self.workspace = dict()
        self._objorder = sasoptpy.util.get_creation_id()
        # TODO Remove '_after' after arranging data/structrues py files
        self._after = False
        self.response = None
        self._internal = False

    def set_response(self, response):
        self.response = response

    def get_response(self):
        return self.response

    def set_internal(self, value):
        self._internal = value

    def is_internal(self):
        return self._internal

    @abstractmethod
    def _defn(self):
        pass

    def _expr(self):
        pass

    @abstractmethod
    def append(self, *args, **kwargs):
        pass



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit test for If-Else-Switch statements.
"""

import os
import sys
import unittest
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '../..')))
from util import assert_equal_wo_temps

from sasoptpy.actions import if_condition, switch_conditions

class TestIfElse(unittest.TestCase):
    """
    Unit tests for IF/ELSE and SWITCH statements
    """

    def setUp(self):
        so.reset()

    def test_regular_if_else(self):
        with so.Workspace('w') as w:
            x = so.Variable(name='x')
            x.set_value(0.5)
            def func1():
                x.set_value(1)
            def func2():
                x.set_value(0)
            if_condition(x > 1e-6, func1, func2)

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x;
               x = 0.5;
               if x > 1e-06 then do;
                  x = 1;
               end;
               else do;
                  x = 0;
               end;
            quit;'''))

    def test_regular_switch(self):
        with so.Workspace('w') as w:
            x = so.Variable(name='x')
            p = so.Parameter(name='p')
            x.set_value(2.5)
            def func1():
                p.set_value(1)
            def func2():
                p.set_value(2)
            def func3():
                p.set_value(3)
            def func4():
                p.set_value(0)

            switch_conditions(x < 1, func1, x < 2, func2, x < 3, func3, func4)

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x;
               num p;
               x = 2.5;
               if x < 1 then do;
                  p = 1;
               end;
               else if x < 2 then do;
                  p = 2;
               end;
               else if x < 3 then do;
                  p = 3;
               end;
               else do;
                  p = 0;
               end;
            quit;'''))

    def test_combined_condition(self):
        with so.Workspace('w') as w:
            p = so.Parameter(name='p')
            def case1():
                p.set_value(10)
            def case2():
                p.set_value(20)
            r = so.Parameter(name='r', value=10)
            if_condition((r < 5) | (r > 10), case1, case2)

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               num p;
               num r = 10;
               if (r < 5) or (r > 10) then do;
                  p = 10;
               end;
               else do;
                  p = 20;
               end;
            quit;'''))



import sasoptpy
import re


def is_abstract(arg):
    return any(isinstance(arg, i) for i in sasoptpy.abstract_classes)


def is_abstract_set(arg):
    return isinstance(arg, sasoptpy.abstract.Set)


def is_key_abstract(arg):
    return isinstance(arg, sasoptpy.abstract.SetIterator) or\
           isinstance(arg, sasoptpy.abstract.SetIteratorGroup)


def is_conditional_value(arg):
    return isinstance(arg, sasoptpy.abstract.Conditional)


def is_solve_statement(i):
    if type(i) == sasoptpy.abstract.SolveStatement:
        return True
    elif isinstance(i, sasoptpy.abstract.Statement):
        if 'solve' in sasoptpy.to_definition(i)[:5]:
            return True
    return False


def is_print_statement(i):
    if type(i) == sasoptpy.abstract.PrintStatement:
        return True
    elif isinstance(i, sasoptpy.abstract.Statement):
        if 'print' in sasoptpy.to_definition(i)[:5]:
            return True
    return False


def is_create_data_statement(i):
    if type(i) == sasoptpy.abstract.CreateDataStatement:
        return True
    return False


def get_key_from_name(name):
    name = name.replace('"', '')
    keys = name.split('[')[1].split(']')[0]
    keys = re.split("(?!\B'[^']*),(?![^']*'\B)", keys)
    keys = [k.replace('\'', '') for k in keys]
    #keys = keys.split(',')
    keys = tuple(int(k) if k.isdigit() else k
                 for k in keys)
    return keys



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
This test file generates intermediate forms to run optimization models using SAS/OR or SAS Viya Optimization solvers.
"""

import os
import sys
import unittest
import tests.examples.responses as expected
import sasoptpy as so
import hashlib

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '../../examples/client_side')))
sys.path.append(os.path.abspath(os.path.join(current_dir, '../../examples/server_side')))
sys.path.append(os.path.abspath(os.path.join(current_dir, '..')))

from util import get_generic_form


class NullWriter:
    """
    Imitates system writer to hide print statements.
    """

    def write(self, text): pass

    def flush(self): pass


class MockCASServer:
    """
    Represents CAS Server for unit tests.
    """

    def __init__(self):
        pass

    def upload_frame(self, *args, **kwargs):
        try:
            return kwargs.get('casout').get('name').upper()
        except IndexError:
            pass


def mock_solve(model, **kwargs):
    """
    Overrides :meth:`Model.solve` method to be used in unittests.

    Parameters
    ----------
    model : :class:`Model` object
        Model to be used in unittest
    kwargs
        Keyword arguments

    Returns
    -------
    boolean
        Success (True) or Failure (False) for the current test
    """
    global ctr
    global expected_response
    global records
    if 'verbose' in kwargs:
        del kwargs['verbose']
    optmodel_code = model.to_optmodel(**kwargs)
    records.append([optmodel_code, expected_response[ctr]])
    ctr += 1
    if so.core.util.is_model(model):
        if model._objval is None:
            model.set_objective_value(0)
    return None


class TestGenerators(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.server = MockCASServer()
        global real_solve
        global real_submit
        global realstdout
        real_solve = so.Model.solve
        real_submit = so.Workspace.submit
        globals()['so'].Model.solve = mock_solve
        globals()['so'].Workspace.submit = mock_solve
        realstdout = sys.stdout
        unittest.util._MAX_LENGTH = 1e+6

    def setUp(self):
        #sys.stdout = NullWriter()
        self.maxDiff = None

    def tearDown(self):
        so.reset()

    def check_results(self):
        sys.stdout = realstdout
        print('Problem:', self.problem)
        for i, solve in enumerate(self.records):
            string0 = get_generic_form(solve[0])
            string1 = get_generic_form(solve[1])
            self.assertMultiLineEqual(string0, string1, string0)
            print('Solve', i, ': True')

    def set_expectation(self, problem, test):
        global ctr
        global expected_response
        global records
        ctr = 0
        expected_response = test
        self.problem = problem
        self.records = records = []


    def test_fm1(self):
        self.set_expectation('Food Manufacture 1', expected.fm1)
        from food_manufacture_1 import test
        test(TestGenerators.server)
        self.check_results()

    def test_fm2(self):
        self.set_expectation('Food Manufacture 2', expected.fm2)
        from food_manufacture_2 import test
        test(TestGenerators.server)
        self.check_results()

    def test_fp1(self):
        self.set_expectation('Factory Planning 1', expected.fp1)
        from factory_planning_1 import test
        test(TestGenerators.server)
        self.check_results()

    def test_fp2(self):
        self.set_expectation('Factory Planning 2', expected.fp2)
        from factory_planning_2 import test
        test(TestGenerators.server)
        self.check_results()

    def test_mp(self):
        self.set_expectation('Manpower Planning', expected.mp)
        from manpower_planning import test
        test(TestGenerators.server)
        self.check_results()

    def test_ro(self):
        self.set_expectation('Refinery Optimization', expected.ro)
        from refinery_optimization import test
        test(TestGenerators.server)
        self.check_results()

    def test_mo(self):
        self.set_expectation('Mining Optimization', expected.mo)
        from mining_optimization import test
        try:
            test(TestGenerators.server)
        except ZeroDivisionError:
            pass
        self.check_results()

    def test_farmp(self):
        self.set_expectation('Farm Planning', expected.farmp)
        from farm_planning import test
        test(TestGenerators.server)
        self.check_results()

    def test_econ(self):
        self.set_expectation('Economic Planning', expected.econ)
        from economic_planning import test
        test(TestGenerators.server)
        self.check_results()

    def test_dc(self):
        self.set_expectation('Decentralization', expected.dc)
        from decentralization import test
        test(TestGenerators.server)
        self.check_results()

    def test_ow(self):
        self.set_expectation('Optimal Wedding', expected.ow)
        from sas_optimal_wedding import test
        test(TestGenerators.server)
        self.check_results()

    def test_kx(self):
        self.set_expectation('Kidney Exchange', expected.kx)
        from sas_kidney_exchange import test
        test(TestGenerators.server)
        self.check_results()

    def test_multiobj(self):
        self.set_expectation('Multiobjective', expected.multiobj)
        from multiobjective import test
        test(TestGenerators.server)
        self.check_results()

    def test_cf(self):
        so.reset()
        self.set_expectation('Curve Fitting', expected.cf)
        from curve_fitting import test
        test(TestGenerators.server)
        self.check_results()

    def test_nl1(self):
        self.set_expectation('Nonlinear 1', expected.nl1)
        from nonlinear_1 import test
        try:
            test(TestGenerators.server)
        except ZeroDivisionError:
            pass
        self.check_results()

    def test_nl2(self):
        self.set_expectation('Nonlinear 2', expected.nl2)
        from nonlinear_2 import test
        test(TestGenerators.server)
        self.check_results()

    def test_least_squares(self):
        self.set_expectation('Least Squares', expected.least_squares)
        from least_squares import test
        test(TestGenerators.server)
        self.check_results()

    def test_efficiency_analysis(self):
        self.set_expectation('Efficiency Analysis', expected.efficiency_analysis)
        from efficiency_analysis import test
        test(TestGenerators.server)
        self.check_results()

    @classmethod
    def tearDownClass(cls):
        global real_solve
        global real_submit
        globals()['so'].Model.solve = real_solve
        globals()['so'].Workspace.submit = real_submit






import warnings

import sasoptpy
from sasoptpy.abstract.util import (
    is_solve_statement, is_print_statement, is_create_data_statement)

class Workspace:
    """
    Workspace represents an OPTMODEL block that allows multiple solves

    Parameters
    ----------
    name : string
        Name of the workspace
    session : :class:`saspy.SASsession` or :class:`swat.cas.connection.CAS`, optional
        Session to be submitted
    """

    def __init__(self, name, session=None):
        self.name = name
        self._load_workspace_defaults()
        self._session = session
        self.response = None
        self._primalSolution = None
        self._dualSolution = None
        self.active_model = '_start_'

    def _load_workspace_defaults(self):
        self._elements = []

    def get_elements(self):
        """
        Returns a list of elements in the workspace
        """
        return self._elements

    def set_active_model(self, model):
        """
        Marks the specified model as active; to be used in solve statements

        Parameters
        ----------
        model : :class:`Model`
            Model to be activated
        """
        self.active_model = model

    def __str__(self):
        return 'Workspace[ID={}]'.format(id(self))

    def __repr__(self):
        return 'sasoptpy.Workspace({})'.format(self.name)

    def __enter__(self):
        sasoptpy.lock.acquire()
        self.original = sasoptpy.container
        sasoptpy.container = self
        return self

    def __exit__(self, type, value, traceback):
        sasoptpy.container = self.original
        sasoptpy.lock.release()

    def append(self, element):
        """
        Appends a new element (operation or statement) to the workspace

        Parameters
        ----------
        element : :class:`sasoptpy.abstract.Statement`
            Any statement that can be appended
        """
        self._elements.append(element)

    def set_session(self, session):
        self._session = session

    def get_session(self):
        return self._session

    def submit(self, **kwargs):
        """
        Submits the workspace as an OPTMODEL block and returns solutions
        """
        return sasoptpy.util.submit(self, **kwargs)

    def parse_solve_responses(self):
        """
        Retrieves the solutions to all solve statements
        """
        keys = self.response.keys()
        solve_statements = [i for i in self.get_elements() if
                            is_solve_statement(i)]

        for i, solve in enumerate(solve_statements):
            ps_key = 'Solve{}.ProblemSummary'.format(i+1)
            ss_key = 'Solve{}.SolutionSummary'.format(i+1)
            ps = None
            ss = None
            if ps_key in keys:
                ps = sasoptpy.interface.parse_optmodel_table(
                    self.response[ps_key])
            if ss_key in keys:
                ss = sasoptpy.interface.parse_optmodel_table(
                    self.response[ss_key])
            if isinstance(solve, sasoptpy.abstract.SolveStatement):
                solve.set_response(problem_summary=ps, solution_summary=ss)
            elif isinstance(solve, sasoptpy.abstract.Statement):
                solve.set_response(ss)

    def parse_print_responses(self):
        """
        Retrieves responses to all print statements
        """
        keys = self.response.keys()
        print_statements = [i for i in self.get_elements() if
                            is_print_statement(i)]

        for i, p in enumerate(print_statements):
            pp_key = 'Print{}.PrintTable'.format(i+1)
            if pp_key in keys:
                if isinstance(p, sasoptpy.abstract.Statement):
                    p.set_response(self.response[pp_key])

    def parse_create_data_responses(self, mediator):
        """
        Grabs responses to all print statements
        """
        keys = self.response.keys()
        cd_statements = [i for i in self.get_elements() if
                            is_create_data_statement(i)]
        session = self._session

        for i, c in enumerate(cd_statements):
            table = c.get_table_expr()
            try:
                tabledf = mediator.parse_table(table)
                c.set_response(tabledf)
            except:
                continue

    def get_variable(self, name):
        """
        Obtains the value of a specified variable name

        Parameters
        ----------
        name : string
            Name of the variable
        """
        vars = filter(
            lambda i: isinstance(i, sasoptpy.Variable), self.get_elements())
        variable = list(filter(lambda i: i.get_name() == name, vars))
        if len(variable) > 1:
            warnings.warn('More than one variable has name {}'.format(name),
                          UserWarning)
        if len(variable) >= 1:
            return variable[0]

        vg = list(filter(
            lambda i: isinstance(i, sasoptpy.VariableGroup) and
                      i.get_name() == name, self.get_elements()))
        if len(vg) > 1:
            warnings.warn(
                'More than one variable group has name {}'.format(name),
                UserWarning)
        if len(vg) >= 1:
            return vg[0]

        if '[' in name:
            group_name = sasoptpy.util.get_group_name(name)
            group = self.get_variable(group_name)
            if group:
                return group.get_member_by_name(name)

        return None

    def set_variable_value(self, name, value):
        """
        Specifies the value of a variable

        Parameters
        ----------
        name : string
            Name of the variable
        value : float
            New value of the variable
        """
        variable = self.get_variable(name)
        if variable is not None:
            variable.set_value(value)

    def to_optmodel(self):
        """
        Returns equivalent OPTMODEL code of the workspace

        Returns
        -------
        optmodel : string
            Generated OPTMODEL code of the workspace object

        """
        return sasoptpy.to_optmodel(self)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit test for structures
"""

import unittest
import sasoptpy as so


class TestStructure(unittest.TestCase):

    def setUp(self):
        so.reset()

    def test_not_implemented(self):
        c = so.statement_dictionary
        so.statement_dictionary = dict()
        from sasoptpy.actions import read_data

        def throw_not_implemented():
            with so.Workspace('w') as w:
                read_data(table='table', index='idx')

        self.assertRaises(NotImplementedError, throw_not_implemented)
        so.util.package_utils.load_function_containers()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit test for create data statement.
"""

import os
import sys
import unittest
import warnings
import sasoptpy as so
from inspect import cleandoc
from sasoptpy.actions import create_data
from sasoptpy.util import concat

from tests.swat_config import create_cas_connection


class TestCreateData(unittest.TestCase):
    """
    Unit tests for CREATE DATA statements
    """

    @classmethod
    def setUpClass(cls):
        so.reset()
        cls.conn = None
        from swat import CAS, SWATError
        try:
            cls.conn = create_cas_connection()
        except SWATError:
            warnings.warn('CAS connection is not available', RuntimeWarning)
        except TypeError:
            warnings.warn('CAS variables are not available', RuntimeWarning)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def test_various_col_types(self):
        # Regular
        with so.Workspace('w') as w:
            m = so.Parameter(name='m', value=7)
            n = so.Parameter(name='n', value=5)
            create_data(table='example', index={}, columns=[m, n])

        self.assertEqual(so.to_optmodel(w), cleandoc('''
        proc optmodel;
           num m = 7;
           num n = 5;
           create data example from m n;
        quit;'''))

        # Column name
        with so.Workspace('w') as w:
            m = so.Parameter(name='m', value=7)
            n = so.Parameter(name='n', value=5)
            create_data(table='example', index={}, columns=[
                {'name': 'ratio', 'expression': m/n}
            ])

        self.assertEqual(so.to_optmodel(w), cleandoc('''
        proc optmodel;
           num m = 7;
           num n = 5;
           create data example from ratio=((m) / (n));
        quit;'''))

        # Custom column name
        with so.Workspace('w') as w:
            m = so.Parameter(name='m', value=7)
            n = so.Parameter(name='n', value=5)
            create_data(table='example', index={}, columns=[
                {'name': concat('s', n), 'expression': m+n}
            ])

        self.assertEqual(so.to_optmodel(w), cleandoc('''
        proc optmodel;
           num m = 7;
           num n = 5;
           create data example from col('s' || n)=(m + n);
        quit;'''))

        # Combined
        with so.Workspace('w') as w:
            m = so.Parameter(name='m', value=7)
            n = so.Parameter(name='n', value=5)
            create_data(table='example', index={}, columns=[
                m, n,
                {'name': 'ratio', 'expression': m/n},
                {'name': concat('s', n), 'expression': m+n}
            ])

        self.assertEqual(so.to_optmodel(w), cleandoc('''
        proc optmodel;
           num m = 7;
           num n = 5;
           create data example from m n ratio=((m) / (n)) col('s' || n)=(m + n);
        quit;'''))

    def test_index(self):
        with so.Workspace('w') as w:
            m = so.ParameterGroup(
                so.exp_range(1, 6), so.exp_range(1, 4), name='m', init=0)
            m[1, 1] = 1
            m[4, 1] = 1
            S = so.Set(name='ISET', value=[i**2 for i in range(1, 3)])
            create_data(
                table='example',
                index={'key': ['i', 'j'], 'set': [S, [1, 2]]},
                columns=[m]
            )

        self.assertEqual(so.to_optmodel(w), cleandoc('''
        proc optmodel;
           num m {1..5, 1..3} init 0;
           m[1, 1] = 1;
           m[4, 1] = 1;
           set ISET = {1,4};
           create data example from [i j] = {{ISET,{1,2}}} m;
        quit;'''))

    def test_regular_index(self):
        with so.Workspace('w') as w:
            s = so.Set(name='S', value=so.exp_range(1, 6))
            x = so.VariableGroup(s, name='x')
            create_data(table='example', index={'key': ['i']}, columns=[x])

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set S = 1..5;
               var x {{S}};
               create data example from [i] x;
            quit;'''))

    def test_custom_index(self):
        with so.Workspace('w') as w:
            s = so.Set(name='S', value=so.exp_range(1, 6))
            x = so.VariableGroup(s, name='x')
            x[1] = 1
            create_data(table='example', index={'key': ['i'], 'set': so.exp_range(1, 4)}, columns=[x])

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set S = 1..5;
               var x {{S}};
               x[1] = 1;
               create data example from [i] = {1..3} x;
            quit;'''))

    def test_subset_index(self):
        with so.Workspace('w') as w:
            s = so.Set(name='S', value=so.exp_range(1, 6))
            x = so.VariableGroup(s, name='x')
            x[1] = 1
            z = so.Set(name='Z', value=[1, 3, 5])
            create_data(table='example', index={'key': ['i'], 'set': z}, columns=[x])

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set S = 1..5;
               var x {{S}};
               x[1] = 1;
               set Z = {1,3,5};
               create data example from [i] = {Z} x;
            quit;'''))

    def test_multi_index(self):
        with so.Workspace('w') as w:
            s = so.Set(name='S', value=so.exp_range(1, 6))
            t = so.Set(name='T', value=so.exp_range(1, 4))
            x = so.VariableGroup(s, t, name='x')
            x[1, 3] = 1
            create_data(
                table='example',
                index={'key': ['i', 'j'], 'set': [s, t]},
                columns=[x])

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set S = 1..5;
               set T = 1..3;
               var x {{S}, {T}};
               x[1, 3] = 1;
               create data example from [i j] = {{S,T}} x;
            quit;'''))

    def test_subset_in_call(self):
        with so.Workspace('w') as w:
            s = so.Set(name='S', value=so.exp_range(1, 6))
            t = so.Set(name='T', value=so.exp_range(1, 4))
            x = so.VariableGroup(s, t, name='x')
            x[1, 3] = 1
            create_data(
                table='example',
                index={'key': ['i', 'j'], 'set': [s, [1, 3]]},
                columns=[x])

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set S = 1..5;
               set T = 1..3;
               var x {{S}, {T}};
               x[1, 3] = 1;
               create data example from [i j] = {{S,{1,3}}} x;
            quit;'''))

    def test_index_with_iterate(self):
        if TestCreateData.conn is None:
            self.skipTest('Session is not available')


        from sasoptpy.util import iterate, concat
        with so.Workspace('w') as w:
            alph = so.Set(name='alph', settype=so.string, value=['a', 'b', 'c'])
            x = so.VariableGroup([1, 2, 3], alph, name='x', init=2)
            with iterate(so.exp_range(1, 4), name='i') as i:
                c = create_data(
                    table='example',
                    index={'key': [i], 'set': [i.get_set()]},
                )

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set <str> alph = {'a','b','c'};
               var x {{1,2,3}, {alph}} init 2;
               create data example from [i] = {{1..3}} ;
            quit;'''))

        # Append column with index
        session = TestCreateData.conn
        with so.Workspace('w', session=session) as w:
            alph = so.Set(name='alph', settype=so.string, value=['a', 'b', 'c'])
            x = so.VariableGroup([1, 2, 3], alph, name='x', init=2)
            with iterate(so.exp_range(1, 4), name='i') as i:
                c = create_data(
                    table='example',
                    index={'key': [i], 'set': [i.get_set()]},
                )
                with iterate(alph, name='j') as j:
                    c.append(
                        {'name': concat('x', j),
                         'expression': x[i, j],
                         'index': j}
                    )

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set <str> alph = {'a','b','c'};
               var x {{1,2,3}, {alph}} init 2;
               create data example from [i] = {{1..3}} {j in alph} < col('x' || j)=(x[i, j]) >;
            quit;'''))

        w.submit()
        self.assertEqual(
            session.CASTable('example').to_frame().to_string(), cleandoc('''
            Selected Rows from Table EXAMPLE
            
                 i   xa   xb   xc
            0  1.0  2.0  2.0  2.0
            1  2.0  2.0  2.0  2.0
            2  3.0  2.0  2.0  2.0'''))

    def test_multiple_iterator(self):
        from sasoptpy.util import concat, iterate
        with so.Workspace('w') as w:
            S = so.Set(name='S', value=[1, 2, 3])
            T = so.Set(name='T', value=[1, 3, 5])
            x = so.VariableGroup(S, T, name='x', init=1)
            with iterate(S, name='i') as i, iterate(T, name='j') as j:
                create_data(
                    table='out',
                    index={},
                    columns=[
                        {'name': concat('x', concat(i, j)),
                         'expression': x[i, j],
                         'index': [i, j]}
                    ]
                )
        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set S = {1,2,3};
               set T = {1,3,5};
               var x {{S}, {T}} init 1;
               create data out from {i in S, j in T} < col('x' || i || j)=(x[i, j]) >;
            quit;'''))

    def test_full_example(self):
        if TestCreateData.conn is None:
            self.skipTest('Session is not available')
        session = TestCreateData.conn
        from sasoptpy.util import concat, iterate

        with so.Workspace('w', session=session) as w:
            m = so.Parameter(name='m', value=3)
            n = so.Parameter(name='n', value=4)
            with iterate(so.exp_range(1, m), 'i') as i, iterate(so.exp_range(1, n), 'j') as j:
                a = so.ParameterGroup(i, j, name='a', value=i * j)
                b = so.ParameterGroup(i, name='b', value=i**2)
            subset = so.Set(name='subset', value=so.exp_range(2, m))
            with iterate(subset, 'i') as i, iterate(so.exp_range(1, n), 'j') as j:
                cd = create_data(
                    table='out',
                    index={'key': [i], 'set': [i.get_set()]},
                    columns=[
                        {'name': concat('a', j), 'expression': a[i, j],
                         'index': j},
                        {'expression': b}
                    ]
                )

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               num m = 3;
               num n = 4;
               num a {i in 1..m, j in 1..n} = i * j;
               num b {i in 1..m} = (i) ** (2);
               set subset = 2..m;
               create data out from [i] = {{subset}} {j in 1..n} < col('a' || j)=(a[i, j]) > b;
            quit;'''))

        w.submit()
        response = session.CASTable('out').to_frame()
        self.assertEqual(response.to_string(), cleandoc('''
            Selected Rows from Table OUT

                 i   a1   a2   a3    a4    b
            0  2.0  2.0  4.0  6.0   8.0  4.0
            1  3.0  3.0  6.0  9.0  12.0  9.0'''))

        response2 = cd.get_response()
        self.assertEqual(response2.to_string(), cleandoc('''
            Selected Rows from Table OUT
            
                 i   a1   a2   a3    a4    b
            0  2.0  2.0  4.0  6.0   8.0  4.0
            1  3.0  3.0  6.0  9.0  12.0  9.0'''))



from .statement_base import Statement

import sasoptpy


class SolveStatement(Statement):

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.model = kwargs.get('model', None)
        self.options = kwargs.get('options', dict())
        self.primalin = kwargs.get('primalin', False)
        self._objorder = sasoptpy.util.get_creation_id()
        self._name = sasoptpy.util.get_next_name()
        self._problem_summary = None
        self._solution_summary = None

    def append(self, element):
        pass

    def _defn(self):

        solve_option_keys = (
            'with', 'obj', 'objective', 'noobj', 'noobjective', 'relaxint',
            'primalin')
        s = ''
        if self.model is not None:
            s = 'use problem {};\n'.format(self.model.get_name())
        s += 'solve'
        pre_opts = []
        pos_opts = []

        options = self.options
        primalin = self.primalin

        if primalin:
            options['primalin'] = True

        if options:
            for key, value in options.items():
                if key in solve_option_keys:
                    if key == 'with':
                        pre_opts.append('with ' + options['with'])
                    elif key == 'relaxint' and options[key] is True:
                        pre_opts.append('relaxint')
                    elif key == 'obj' or key == 'objectives':
                        pre_opts.append('obj ({})'.format(
                            ' '.join(i.get_name() for i in options[key])))
                        multi_obj = True
                    elif key == 'primalin' and options[key] is True:
                        pos_opts.append('primalin')
                else:
                    if type(value) is dict:
                        pos_opts.append('{}=('.format(key) + ','.join(
                            '{}={}'.format(i, j)
                            for i, j in value.items()) + ')')
                    else:
                        pos_opts.append('{}={}'.format(key, value))

            if pre_opts:
                s += ' ' + ' '.join(pre_opts)
            if pos_opts:
                s += ' / ' + ' '.join(pos_opts)
        s += ';'
        return s

    def set_response(self, problem_summary, solution_summary):
        self._problem_summary = problem_summary
        self._solution_summary = solution_summary
        self.response = {
            'Problem Summary': self._problem_summary,
            'Solution Summary': self._solution_summary
        }

    def get_problem_summary(self):
        return self._problem_summary

    def get_solution_summary(self):
        return self._solution_summary

    @classmethod
    def solve(cls, *args, **kwargs):
        st = SolveStatement(*args, **kwargs)
        return st

    @classmethod
    def solve_model(cls, model, **kwargs):
        kwargs['model'] = model
        st = SolveStatement(**kwargs)
        return st



import sasoptpy
from sasoptpy.libs import pd
from abc import ABC, abstractmethod
from collections import OrderedDict


class Group(ABC):

    def __init__(self, name):
        self._name = name
        self._groups = OrderedDict()

    def get_name(self):
        return self._name

    def _register_keys(self, keys):
        for j, k in enumerate(keys):
            try:
                self._groups[j].append(k)
            except KeyError:
                self._groups[j] = [k]

    def filter_unique_keys(self):
        for i in self._groups:
            self._groups[i] = list(OrderedDict.fromkeys(self._groups[i]))

    def get_group_types(self):
        group_types = []
        for i in self._groups.values():
            if any([sasoptpy.abstract.is_key_abstract(j) for j in i]):
                group_types.append(sasoptpy.ABSTRACT)
            else:
                group_types.append(sasoptpy.CONCRETE)
        return group_types

    @abstractmethod
    def get_members(self):
        """
        Each Group object should provide a way to access its members as an
        OrderedDict
        """



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit test for configurations
"""

import unittest
import sasoptpy as so


class TestConfig(unittest.TestCase):

    def setUp(self):
        self.define_problem()

    def define_problem(self):
        self.model = so.Model(name='model_test_config')
        self.var = self.model.add_variable(name='x')
        self.con = self.model.add_constraint(
            10 / 3 * self.var + 1e-20 * self.var ** 2 <= 30 + 1e-11,
            name='c')
        self.target_config = 'max_digits'
        self.expected_response = {
            2: "con c : 3.33 * x + 0.0 * ((x) ^ (2)) <= 30.0;",
            3: "con c : 3.333 * x + 0.0 * ((x) ^ (2)) <= 30.0;",
            None: "con c : 3.3333333333333335 * x + 1e-20 * ((x) ^ (2)) <= 30.00000000001;"
        }
        self.new_config_key = 'new_config'
        self.new_config_value = 20

    def measure_function(self):
        return self.con._defn()

    def test_default_configs(self):
        self.assertEqual(set(so.default_config_keys), set(so.config.keys))

    def test_config_constructor(self):
        for i in self.expected_response.keys():
            so.config = so.Config(self.target_config, i)
            actual_response = self.measure_function()
            self.assertEqual(actual_response, self.expected_response[i])

    def test_override_config(self):
        for i in self.expected_response.keys():
            so.config[self.target_config] = i
            actual_response = self.measure_function()
            self.assertEqual(actual_response, self.expected_response[i])

    def test_delete_overridden(self):
        original_value = so.config[self.target_config]
        so.config[self.target_config] = original_value + 1
        del so.config[self.target_config]
        self.assertEqual(so.config[self.target_config], original_value)

    def test_config_reset(self):
        original_value = so.config[self.target_config]
        so.config[self.target_config] = original_value + 1
        so.config.reset()
        self.assertEqual(so.config[self.target_config], original_value)

    def test_add_delete_config(self):
        so.config[self.new_config_key] = self.new_config_value
        self.assertEqual(so.config[self.new_config_key], self.new_config_value)
        del so.config[self.new_config_key]
        self.assertEqual(so.config[self.new_config_key], None)

    def test_delete_key_exception(self):
        def delete_nonexistent_key():
            del so.config['nonexistent_key']
        self.assertRaises(KeyError, delete_nonexistent_key)

    def test_config_iter(self):
        list_of_keys_via_iter = set()
        list_of_keys_via_keys = set()

        for i in so.config:
            list_of_keys_via_iter.add(i)
        for i in so.config.keys:
            list_of_keys_via_keys.add(i)

        self.assertEqual(list_of_keys_via_iter, list_of_keys_via_keys)

    def tearDown(self):
        so.reset()




from .set_iterator import SetIterator, SetIteratorGroup
from .statement.statement_base import Statement
import sasoptpy
from types import GeneratorType


class Set:
    """
    Creates an index set to be represented inside PROC OPTMODEL

    Parameters
    ----------
    name : string
        Name of the parameter
    init : :class:`Expression`, optional
        Initial value expression of the parameter
    settype : list, optional
        List of types for the set, consisting of 'num' and 'str' values

    Examples
    --------

    >>> I = so.Set('I')
    >>> print(I._defn())
    set I;

    >>> J = so.Set('J', settype=['num', 'str'])
    >>> print(J._defn())
    set <num, str> J;

    >>> N = so.Parameter(name='N', init=5)
    >>> K = so.Set('K', init=so.exp_range(1,N))
    >>> print(K._defn())
    set K = 1..N;

    """

    @sasoptpy.class_containable
    def __init__(self, name, init=None, value=None, settype=None):
        self._name = name
        if name is not None:
            self._objorder = sasoptpy.util.get_creation_id()
        self._init = init
        self._value = value
        if settype is None:
            settype = ['num']
        self._type = sasoptpy.util.pack_to_list(settype)
        self._colname = sasoptpy.util.pack_to_list(name)
        self._iterators = []

    @classmethod
    def from_object(cls, obj):
        return Set(name=None, value=obj)

    def get_name(self):
        return self._name

    def __iter__(self):
        if len(self._type) > 1:
            s = SetIteratorGroup(self, datatype=self._type)
            self._iterators.append(s)
            return iter([s])
        else:
            s = SetIterator(self)
            self._iterators.append(s)
            return iter([s])

    def _defn(self):
        s = 'set '
        if len(self._type) == 1 and self._type[0] == 'num':
            s += ''
        else:
            s += '<' + ', '.join(self._type) + '> '
        s += self.get_name()
        if self._init is not None:
            s += ' init ' + sasoptpy.util.package_utils._to_sas_string(self._init) #str(self._init)
        elif self._value is not None:
            s += ' = ' + sasoptpy.util.package_utils._to_sas_string(self._value)
        s += ';'
        return(s)

    def __hash__(self):
        return hash(self.get_name())

    def __contains__(self, item):
        from .util import is_conditional_value
        if is_conditional_value(item):
            return item.is_in(self)
        else:
            raise RuntimeError('Cannot verify if value is in abstract set')

    def __str__(self):
        s = self.get_name()
        return('{}'.format(s))

    def __repr__(self):
        s = 'sasoptpy.abstract.Set(name={}, settype={}'.format(
            self.get_name(), self._type)
        if self._init:
            s += ', init={}'.format(self._init)
        if self._value:
            s += ', value={}'.format(self._value)
        s += ')'
        return(s)

    def _expr(self):
        if self.get_name() is not None:
            return self.get_name()
        else:
            return sasoptpy.to_expression(self._value)


class InlineSet(Statement):

    def __init__(self, func):
        super().__init__()
        self.sym = sasoptpy.abstract.Conditional(self)
        with sasoptpy.set_container(self, conditions=True):
            res = func()
            for m in res:
                if isinstance(m, tuple):
                    for j in m:
                        self.append(j)
                else:
                    self.append(m)

    def append(self, element):
        self.elements.append(element)

    def _defn(self):
        pass

    def _expr(self):
        member_defs = ', '.join(sasoptpy.to_definition(i) for i in self.elements)
        conditions = self.sym.get_conditions_str()
        if conditions != '':
            return f'{{{member_defs}: {conditions}}}'
        else:
            return f'{{{member_defs}}}'

    def __repr__(self):
        s = 'sasoptpy.InlineSet('
        try:
            s += self._expr()
        except:
            s += f'id={id(self)}'
        s += '}'
        return s



import sasoptpy as so
import pandas as pd


def test(cas_conn):

    m = so.Model(name='economic_planning', session=cas_conn)

    industry_data = pd.DataFrame([
        ['coal', 150, 300, 60],
        ['steel', 80, 350, 60],
        ['transport', 100, 280, 30]
        ], columns=['industry', 'init_stocks', 'init_productive_capacity',
                    'demand']).set_index(['industry'])

    production_data = pd.DataFrame([
        ['coal', 0.1, 0.5, 0.4],
        ['steel', 0.1, 0.1, 0.2],
        ['transport', 0.2, 0.1, 0.2],
        ['manpower', 0.6, 0.3, 0.2],
        ], columns=['input', 'coal',
                    'steel', 'transport']).set_index(['input'])

    productive_capacity_data = pd.DataFrame([
        ['coal', 0.0, 0.7, 0.9],
        ['steel', 0.1, 0.1, 0.2],
        ['transport', 0.2, 0.1, 0.2],
        ['manpower', 0.4, 0.2, 0.1],
        ], columns=['input', 'coal',
                    'steel', 'transport']).set_index(['input'])

    manpower_capacity = 470
    num_years = 5

    YEARS = list(range(1, num_years+1))
    YEARS0 = [0] + list(YEARS)
    INDUSTRIES = industry_data.index.tolist()

    init_stocks = industry_data['init_stocks']
    init_productive_capacity = industry_data['init_productive_capacity']
    demand = industry_data['demand']

    production_coeff = so.flatten_frame(production_data)
    productive_capacity_coeff = so.flatten_frame(productive_capacity_data)

    static_production = m.add_variables(INDUSTRIES, lb=0,
                                        name='static_production')
    m.set_objective(0, sense=so.MIN, name='Zero')
    m.add_constraints((static_production[i] == demand[i] +
                       so.expr_sum(
                           production_coeff[i, j] * static_production[j]
                           for j in INDUSTRIES) for i in INDUSTRIES),
                      name='static_con')

    m.solve()
    print(so.get_value_table(static_production))

    final_demand = so.get_value_table(
        static_production)['static_production']

    production = m.add_variables(INDUSTRIES, range(0, num_years+2), lb=0,
                                 name='production')
    stock = m.add_variables(INDUSTRIES, range(0, num_years+2), lb=0,
                            name='stock')
    extra_capacity = m.add_variables(INDUSTRIES, range(2, num_years+3), lb=0,
                                     name='extra_capacity')

    productive_capacity = so.ImplicitVar(
        (init_productive_capacity[i] +
         so.expr_sum(extra_capacity[i, y] for y in range(2, year+1))
         for i in INDUSTRIES for year in range(1, num_years+2)),
        name='productive_capacity'
    )

    for i in INDUSTRIES:
        production[i, 0].set_bounds(ub=0)
        stock[i, 0].set_bounds(lb=init_stocks[i], ub=init_stocks[i])

    total_productive_capacity = sum(productive_capacity[i, num_years]
                                    for i in INDUSTRIES)
    total_production = so.expr_sum(production[i, year] for i in INDUSTRIES
                                    for year in [4, 5])
    total_manpower = so.expr_sum(production_coeff['manpower', i] *
                                  production[i, year+1] +
                                  productive_capacity_coeff['manpower', i] *
                                  extra_capacity[i, year+2]
                                  for i in INDUSTRIES for year in YEARS)

    continuity_con = m.add_constraints((
        stock[i, year] + production[i, year] ==
        (demand[i] if year in YEARS else 0) +
        so.expr_sum(production_coeff[i, j] * production[j, year+1] +
                     productive_capacity_coeff[i, j] *
                     extra_capacity[j, year+2] for j in INDUSTRIES) +
        stock[i, year+1]
        for i in INDUSTRIES for year in YEARS0), name='continuity_con')

    manpower_con = m.add_constraints((
        so.expr_sum(production_coeff['manpower', j] * production[j, year] +
                     productive_capacity_coeff['manpower', j] *
                     extra_capacity[j, year+1]
                     for j in INDUSTRIES)
        <= manpower_capacity for year in range(1, num_years+2)),
        name='manpower_con')

    capacity_con = m.add_constraints((production[i, year] <=
                                      productive_capacity[i, year]
                                      for i in INDUSTRIES
                                      for year in range(1, num_years+2)),
                                     name='capacity_con')

    for i in INDUSTRIES:
        production[i, num_years+1].set_bounds(lb=final_demand[i])

    for i in INDUSTRIES:
        for year in [num_years+1, num_years+2]:
            extra_capacity[i, year].set_bounds(ub=0)

    problem1 = so.Model(name='Problem1', session=cas_conn)
    problem1.include(
        production, stock, extra_capacity, continuity_con, manpower_con,
        capacity_con, productive_capacity)
    problem1.set_objective(total_productive_capacity, sense=so.MAX,
                           name='total_productive_capacity')
    problem1.solve()
    so.pd.display_dense()
    print(so.get_value_table(production, stock, extra_capacity,
                                productive_capacity).sort_index())
    print(so.get_value_table(manpower_con.get_expressions()))

    # Problem 2

    problem2 = so.Model(name='Problem2', session=cas_conn)
    problem2.include(problem1)
    problem2.set_objective(total_production, name='total_production',
                           sense=so.MAX)
    for i in INDUSTRIES:
        for year in YEARS:
            continuity_con[i, year].set_rhs(0)
    problem2.solve()
    print(so.get_value_table(production, stock, extra_capacity,
                                productive_capacity).sort_index())
    print(so.get_value_table(manpower_con.get_expressions()))

    # Problem 3

    problem3 = so.Model(name='Problem3', session=cas_conn)
    problem3.include(production, stock, extra_capacity, continuity_con,
                     capacity_con)
    problem3.set_objective(total_manpower, sense=so.MAX, name='total_manpower')
    for i in INDUSTRIES:
        for year in YEARS:
            continuity_con[i, year].set_rhs(demand[i])
    problem3.solve()
    print(so.get_value_table(production, stock, extra_capacity,
                                productive_capacity).sort_index())
    print(so.get_value_table(manpower_con.get_expressions()))

    return problem3.get_objective_value()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit test for abstract assignments.
"""

import os
import sys
import unittest
import warnings
import sasoptpy as so
from inspect import cleandoc
from sasoptpy.actions import read_data

from tests.swat_config import create_cas_connection


class TestAssignment(unittest.TestCase):
    """
    Unit tests for assignment statements
    """

    @classmethod
    def setUpClass(cls):
        so.reset()
        cls.conn = None
        from swat import CAS, SWATError
        try:
            cls.conn = create_cas_connection()
        except SWATError:
            warnings.warn('CAS connection is not available', RuntimeWarning)
        except TypeError:
            warnings.warn('CAS variables are not available', RuntimeWarning)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def test_bound_assignment(self):

        with so.Workspace('test_regular_assignment') as w:

            p = so.Parameter(name='p', init=2)
            x = so.VariableGroup(5, lb=1, name='x')
            x[0].set_bounds(lb=3)
            x[1].set_bounds(lb=p)
            p.set_value(4)
            x[2].set_bounds(lb=p, ub=p)
            x[3].set_bounds(ub=5)
            x[4].set_bounds(lb=1, ub=4)

        self.assertEqual(so.to_optmodel(w), cleandoc("""
            proc optmodel;
               num p init 2;
               var x {{0,1,2,3,4}} >= 1;
               x[0].lb = 3;
               x[1].lb = p;
               p = 4;
               fix x[2]=p;
               x[3].ub = 5;
               x[4].lb = 1;
               x[4].ub = 4;
            quit;
            """))

    def test_assignment_append(self):
        x = so.Variable(name='x')
        r = so.abstract.Assignment(x, 5)
        self.assertEqual(so.to_definition(r), 'x = 5;')
        y = so.Variable(name='y')
        r.append(identifier=y)
        self.assertEqual(so.to_definition(r), 'y = 5;')
        r.append(expression=10)
        self.assertEqual(so.to_definition(r), 'y = 10;')
        r.append(keyword='fix')
        self.assertEqual(so.to_definition(r), 'fix y = 10;')

    def test_fix_value(self):
        from sasoptpy.actions import fix
        with so.Workspace('w') as w:
            e = so.Parameter(name='e', value=4)
            x = so.VariableGroup(5, name='x')
            fix(x[0], 0)
            fix(x[1], e*2)
        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               num e = 4;
               var x {{0,1,2,3,4}};
               fix x[0]=0;
               fix x[1]=2 * e;
            quit;'''))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit test for drop statements.
"""

import os
import sys
import unittest
import warnings
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '../..')))
from util import assert_equal_wo_temps

class TestDropRestore(unittest.TestCase):
    """
    Unit tests for DROP and RESTORE statements
    """

    def setUp(self):
        so.reset()

    def test_regular_drop(self):

        import sasoptpy.abstract.math as sm
        from sasoptpy.actions import solve, drop

        with so.Workspace('w') as w:
            x = so.Variable(name='x', lb=1)
            y = so.Variable(name='y', lb=0)
            c = so.Constraint(sm.sqrt(x) >= 5, name='c')
            o = so.Objective(x + y, sense=so.MIN, name='obj')
            s = solve()
            drop(c)
            o2 = so.Objective(x, sense=so.MIN, name='obj2')
            s2 = solve()

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x >= 1;
               var y >= 0;
               con c : sqrt(x) >= 5;
               min obj = x + y;
               solve;
               drop c;
               min obj2 = x;
               solve;
            quit;'''))

    def test_multiple_and_mixed_drop(self):

        from sasoptpy.actions import solve, drop
        with so.Workspace('w') as w:
            x = so.Variable(name='x')
            o = so.Objective(x**2, sense=so.MIN, name='obj')
            c1 = so.Constraint(x >= 5, name='c1')
            c2 = so.ConstraintGroup((x**i >= 1+i for i in range(3)), name='c2')
            c3 = so.ConstraintGroup((x >= i for i in range(2)), name='c3')
            S = so.Set(name='S', value=range(1, 5))
            y = so.VariableGroup(S, name='y')
            d = so.ConstraintGroup((y[i] >= i+j for i in S for j in range(2)),
                                   name='d')
            solve()
            drop(c1, c2[1], c3)
            solve()
            drop(d[2,0])
            solve()
            drop(d)
            solve()

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x;
               min obj = (x) ^ (2);
               con c1 : x >= 5;
               con c2_0 : (x) ^ (0) >= 1;
               con c2_1 : (x) ^ (1) >= 2;
               con c2_2 : (x) ^ (2) >= 3;
               con c3_0 : x >= 0;
               con c3_1 : x >= 1;
               set S = 1..4;
               var y {{S}};
               con d_0 {TEMP1 in S} : y[TEMP1] - TEMP1 >= 0;
               con d_1 {TEMP1 in S} : y[TEMP1] - TEMP1 >= 1;
               solve;
               drop c1 c2_1 c3_0 c3_1;
               solve;
               drop d_0[2];
               solve;
               drop {TEMP1 in S} d_0[TEMP1] {TEMP1 in S} d_1[TEMP1];
               solve;
            quit;'''))

    def test_drop_in_model(self):

        m = so.Model(name='m')
        x = m.add_variables(3, name='x')
        c = m.add_constraints((x[i] <= i**2 for i in range(3)), name='c')
        self.assertEqual(so.to_optmodel(m), cleandoc('''
            proc optmodel;
               min m_obj = 0;
               var x {{0,1,2}};
               con c_0 : x[0] <= 0;
               con c_1 : x[1] <= 1;
               con c_2 : x[2] <= 4;
               solve;
            quit;'''))
        m.drop_constraint(c[0])
        self.assertEqual(so.to_optmodel(m), cleandoc('''
            proc optmodel;
               min m_obj = 0;
               var x {{0,1,2}};
               con c_0 : x[0] <= 0;
               con c_1 : x[1] <= 1;
               con c_2 : x[2] <= 4;
               drop c_0;
               solve;
            quit;'''))

    def test_simple_drop_restore(self):
        from sasoptpy.actions import drop, restore, solve, set_objective
        with so.Workspace('w') as w:
            x = so.Variable(name='x', lb=-1)
            set_objective(x**3, name='xcube', sense=so.minimize)
            c = so.Constraint(x >= 1, name='xbound')
            solve()
            drop(c)
            solve()
            restore(c)
            solve()

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x >= -1;
               MIN xcube = (x) ^ (3);
               con xbound : x >= 1;
               solve;
               drop xbound;
               solve;
               restore xbound;
               solve;
            quit;'''))
